"""NextGenWindow module."""

from logging import Logger
from t_nextgen.ng_app_manager import NGAppManager


class NextGenWindow:
    """NextGenWindow class."""

    def __init__(self, app_path: str, logger: Logger) -> None:
        """Initialize NextGenWindow."""
        self.desktop_app = NGAppManager(app_path, logger)
        self.logger = logger



import contextlib
import _ctypes
import time

from pywinauto.application import WindowSpecification

from t_nextgen.config import Config
from t_nextgen.ng_pm.core import NextGenPMCore
from t_nextgen.utils.logger import logger


class BatchDeleter:
    """Class to delete batches from the NextGen PM application."""

    def __init__(
        self,
        next_gen: NextGenPMCore,
        descriptions: list[str],
        practice: str,
    ):
        """Initialize the BatchDeleter object.

        Args:
            next_gen (NextGenPMCore): NextGen PM Core object.
            descriptions (list[str]): List of descriptions of the batches to delete.
            practice (str): Practice name.
        """
        self.next_gen = next_gen
        self.descriptions = descriptions
        self.practice = practice
        self.batches_not_secured_to_thoughtful: list[str] = []
        self.number_of_remits = 0
        self.found_batch_to_delete = False
        self.batches_marked_to_deletion: list[str] = []
        self.batches_not_found_for_deletion: list[str] = []
        self.total_number_of_batches_in_the_practice = 0
        self.batches_not_deleted: list[str] = []
        self.all_batches_secured_to_thoughtful_were_deleted: bool = False

    def get_batch_rows(self) -> WindowSpecification:
        """Get the batch rows from the batch posting window."""
        pane = self.next_gen.desktop_app.dialog.child_window(title="lstListing", control_type="Pane")
        return pane.descendants(control_type="DataItem")

    def mark_batch_for_deletion(self, data_item: WindowSpecification, description: str):
        """Mark the batch for deletion.

        Args:
            data_item (WindowSpecification): DataItem object of the batch.
            description (str): Description of the batch.
        """
        logger.info(f"Selecting batch for deletion: {description}")
        with contextlib.suppress(_ctypes.COMError):
            data_item.descendants(control_type="CheckBox")[0].toggle()

    def click_delete_option(self):
        """Click the delete option in the batch posting window."""
        logger.info(f"Deleting {self.number_of_remits} remits from {self.practice}")
        self.next_gen.batch_posting_window.click_menu_icon("d")
        time.sleep(2)
        with contextlib.suppress(_ctypes.COMError):
            self.next_gen.desktop_app.dialog.child_window(title="OK", control_type="Button").click()

    def wait_for_batches_to_be_deleted(self, timeout: int = 300, interval: int = 30) -> None:
        """Waits for the batches to be deleted.

        Args:
            timeout (int): Maximum time to wait for deletion in seconds. Defaults to 300 seconds.
            interval (int): Time interval between checks in seconds. Defaults to 10 seconds.
        """
        logger.info("Waiting for batches to be deleted")
        start_time = time.time()
        while time.time() - start_time < timeout:
            try:
                found_batch = False
                for data_item in self.get_batch_rows():
                    description = data_item.descendants(title="Description", control_type="Edit")[0].get_value()
                    if any(desc in description for desc in self.batches_marked_to_deletion):
                        found_batch = True
                        break
                if not found_batch:
                    logger.info("All batches have been deleted.")
                    return
                logger.info("Batches still found. Waiting...")
                time.sleep(interval)
            except RuntimeError:
                logger.warning("NextGen app seems to be in a freeze state. Waiting for it to recover.")
                time.sleep(10)
        logger.warning("Timeout reached. Some batches may not have been deleted.")

    def check_if_some_batch_was_not_found_for_deletion(self):
        """Check if some batches were not found for deletion."""
        self.batches_not_found_for_deletion = [
            batch
            for batch in self.descriptions
            if (batch not in self.batches_not_secured_to_thoughtful and batch not in self.batches_marked_to_deletion)
        ]
        if self.batches_not_found_for_deletion:
            logger.info(
                f"Some batches were not found for deletion in {self.practice}."
                f"Batches not found: {self.batches_not_found_for_deletion}"
            )

    def check_if_all_batches_were_deleted(self):
        """Check if all batches were deleted."""
        data_items = self.get_batch_rows()
        if self.batches_marked_to_deletion and (
            self.total_number_of_batches_in_the_practice - len(self.batches_not_secured_to_thoughtful)
            == len(data_items)
        ):
            logger.info(f"All batches secured to thoughtful were deleted in {self.practice}")
            self.all_batches_secured_to_thoughtful_were_deleted = True
        else:
            batches_secured_to_thoughtful = [
                batch for batch in self.descriptions if batch not in self.batches_not_secured_to_thoughtful
            ]

            for index, data_item in enumerate(data_items):
                next_gen_description = data_item.descendants(title="Description", control_type="Edit")[0].get_value()
                for description in batches_secured_to_thoughtful:
                    if description in next_gen_description:
                        self.batches_not_deleted.append(description)
                        break
            if self.batches_not_deleted:
                logger.warning(
                    f"Some batches were not deleted in {self.practice}. Batches not deleted: {self.batches_not_deleted}"
                )
                self.all_batches_secured_to_thoughtful_were_deleted = False

    def check_if_all_batches_were_selected_correctly(self):
        """Check if all batches were selected correctly."""
        batch_rows = self.get_batch_rows()
        batch_selected = 0
        for data_item in batch_rows:
            check_box = data_item.descendants(control_type="CheckBox")[0]
            if check_box.get_toggle_state() == 1:
                batch_selected += 1
        if batch_selected == len(self.batches_marked_to_deletion):
            logger.info("All batches were selected correctly for deletion.")
        else:
            logger.warning("Some batches were not selected correctly for deletion.")

    def generate_report(self):
        """Generate a report of the batches that were not deleted."""
        return {
            "practice": self.practice,
            "batches_not_secured_to_thoughtful": self.batches_not_secured_to_thoughtful,
            "number_of_batches_marked_to_deletion": len(self.batches_marked_to_deletion),
            "batches_not_found_for_deletion": self.batches_not_found_for_deletion,
            "batches_not_deleted": self.batches_not_deleted,
            "number_of_batches_to_delete": len(self.descriptions),
            "number_of_batches_not_found_for_deletion": len(self.batches_not_found_for_deletion),
            "number_of_batches_secured_to_thoughtful": len(self.batches_marked_to_deletion),
            "number_of_batches_not_secured_to_thoughtful": len(self.batches_not_secured_to_thoughtful),
            "number_of_batches_in_the_practice": self.total_number_of_batches_in_the_practice,
            "number_of_remits_deleted": self.number_of_remits,
            "all_batches_secured_to_thoughtful_deleted": self.all_batches_secured_to_thoughtful_were_deleted,
        }

    def delete(self) -> dict:
        """Delete the batches."""
        self.next_gen.batch_posting_window.click_batch_icon_from_bar(self.practice)
        data_items = self.get_batch_rows()
        self.total_number_of_batches_in_the_practice = len(data_items)
        for index, data_item in enumerate(data_items):
            next_gen_description = data_item.descendants(title="Description", control_type="Edit")[0].get_value()
            secured_to = data_item.descendants(title="Secured To", control_type="Edit")[0].get_value()
            for description in self.descriptions:
                if description in next_gen_description:
                    if not data_item.is_visible():
                        self.next_gen.desktop_app.click_down_n_times(index)

                    if "thoughtful" in secured_to.lower():
                        members = data_item.descendants(title="Members", control_type="Edit")[0].get_value()
                        self.number_of_remits += int(members)
                        self.found_batch_to_delete = True
                        self.mark_batch_for_deletion(data_item, description)
                        self.batches_marked_to_deletion.append(description)
                    else:
                        self.batches_not_secured_to_thoughtful.append(description)
                        logger.warning(f"Batch {description} is secured to {secured_to}. Not marking it to delete")
                    break
        if self.found_batch_to_delete:
            self.check_if_all_batches_were_selected_correctly()
            self.click_delete_option()
            self.wait_for_batches_to_be_deleted(timeout=self.number_of_remits * 40, interval=40)
        else:
            logger.info(f"No batches found to delete in {self.practice}")
        self.check_if_all_batches_were_deleted()
        self.check_if_some_batch_was_not_found_for_deletion()
        return self.generate_report()


def delete_batches(
    next_gen: NextGenPMCore, batches_to_delete: list[dict], database: str = Config.DATABASES.TEST
) -> list[dict]:
    """Deletes batches from the NextGen PM application.

    Args:
        next_gen (NextGenPMCore): NextGen PM Core object.
        batches_to_delete (list[dict]): List of dictionaries containing practice name and descriptions to delete.
        database (str): Database name. Options: "NGPROD" or "NGTEST". Defaults to "NGTEST".

    Example:
        batches_to_delete =[
            {
                "practice": "Proliance Southwest Seattle Orthopedics",
                "descriptions": [
                    "*****",
                    "******",
                    "******",
                ]
            },
            {
                "practice": "Proliance Hand Wrist & Elbow Physicians",
                "descriptions": [
                    "*******",
                    "*******",
                    "*****",
                ]
            }
        ]
    """
    report_list = []
    next_gen.login(practice="_Central Office", database=database)
    next_gen.desktop_app.click_no_button_in_next_gen_alert()
    for batches in batches_to_delete:

        logger.info(f"Selecting practice: {batches['practice']}")
        try:
            next_gen.select_practice_from_app(batches["practice"])
        except Exception as e:
            logger.warning(f"Error selecting practice {batches['practice']}. Error: {e}")
            # todo need to click cancel in the change practice popup
            continue
        batch_deleter = BatchDeleter(next_gen, batches["descriptions"], batches["practice"])
        report = batch_deleter.delete()
        report_list.append(report)

    if next_gen_process := next_gen.desktop_app.get_app_session_if_running(next_gen.app_path):
        logger.debug("Closing the NextGen process.")
        next_gen.close_session(next_gen_process)
    return report_list


if __name__ == "__main__":
    batches_to_delete = [
        {
            "practice": "Proliance Southwest Seattle Orthopedics",
            "descriptions": ["*******", "*******", "********"],
        },
    ]
    next_gen = NextGenPMCore()
    delete_batches(next_gen, batches_to_delete)



from decimal import Decimal
from typing import Any
import re


def convert_to_decimal(str_value: Any, empty: bool = False) -> Decimal:
    """The function `convert_to_decimal` takes a string input, removes non-numeric characters.

    Returns a Decimal value, handling exceptions by returning 0.0 if conversion fails.

    Returns:
        Decimal: Returns a Decimal value after converting the input `str_value` to a Decimal
    """
    if empty and not str_value:  # there are places where we need an empty value instead of turn it to 0.0
        return str_value
    if isinstance(str_value, str):
        str_value = str_value.replace(",", "")
        # This regex will match numbers inside () -> we are getting it in the qty/charge
        parenthesis_pattern = r"\(([-+]?\d+)\)"
        dollar_pattern = r"\$"
        combined_pattern = f"{parenthesis_pattern}|{dollar_pattern}"
        str_value = re.sub(combined_pattern, "", str_value)
        str_value = str_value.translate(str.maketrans("", "", "()"))
        str_value = str_value.replace(" ", "")
        if str_value == "":
            return Decimal("0.0")
    return Decimal(str(str_value))



"""Module for BatchMaintenance class."""
import contextlib
import re

from t_desktop.t_desktop import PywinTimeout

from t_nextgen.nextgen_window import NextGenWindow

import _ctypes
from t_desktop.config import IS_WINDOWS_OS

if IS_WINDOWS_OS:
    from pywinauto.controls.uia_controls import ComboBoxWrapper
    from pywinauto.findwindows import ElementNotFoundError
    from pywinauto.timings import TimeoutError


class BatchMaintenanceWindow(NextGenWindow):
    """Batch Maintenance class with methods to interact with batch maintenance window."""

    @property
    def window(self):
        """Return the AMBatches window element."""
        return self.desktop_app.dialog.child_window(title="AMBatches", control_type="Window")

    def click_cancel(self) -> None:
        """Clicks the cancel button on the maintenance window and confirms by clicking 'No'."""
        with contextlib.suppress(_ctypes.COMError):
            self.window.child_window(auto_id="cmdCancel", control_type="Button").click()
        self.logger.debug("Clicked Cancel button on Batch Maintenance window.")
        self.click_no_on_popup_if_exist()

    def click_no_on_popup_if_exist(self) -> None:
        """Clicks the 'No' button on the maintenance window pop-up if it appears."""
        try:
            self.desktop_app.wait_until_element_visible(title="NextGen", control_type="Window")
            next_gen_popup = self.desktop_app.dialog.child_window(title="NextGen", control_type="Window")
            no_button = next_gen_popup.child_window(auto_id="7", control_type="Button")
            with contextlib.suppress(_ctypes.COMError):
                no_button.click()
            no_button.wait_not("visible", timeout=10)
            self.logger.debug("Clicked 'No' on NextGen pop-up.")
        except (ElementNotFoundError, TimeoutError):
            self.logger.debug("Save confirmation pop-up not found; proceeding without further action.")

    def click_ok(self) -> None:
        """Clicks the OK button and waits for it to disappear."""
        ok_button = self.window.child_window(auto_id="cmdOK", control_type="Button")
        with contextlib.suppress(_ctypes.COMError):
            ok_button.click()
        self.desktop_app.close_modal(buttons_to_try=["YES"])
        ok_button.wait_not("visible", timeout=10)
        self.logger.debug("Clicked OK on Batch Maintenance window.")

    def get_desc_amount(self) -> str:
        """Fetch Batch Desc Amount from Batch Maintenance Window.

        Returns:
            str: desc amount
        """
        amt = self.window.child_window(auto_id="txtDesc", control_type="Edit").get_value()
        return re.search(r"(-?\$-?[0-9]+\.?[0-9]*)", amt).group().replace("$", "")

    def update_total_billed_header(self, ledger_billed: str) -> None:
        """Update Billed Amount Header in the Batch Header Window.

        Args:
            ledger_billed (str): The amount billed to be updated in the batch header.
        """
        self.logger.debug(f"Updating Billed Amount Header in the Batch Header Window to {ledger_billed}.")
        with contextlib.suppress(_ctypes.COMError):
            self.window.child_window(auto_id="txtTotBilled").set_text(str(ledger_billed))

    def update_total_allowed_header(self, ledger_allowed: str) -> None:
        """Update Allowed Amount Header in the Batch Header Window.

        Args:
            ledger_allowed (str): The allowed amount to be updated in the batch header.
        """
        self.logger.debug(f"Updating Allowed Amount Header in the Batch Header Window to {ledger_allowed}.")
        with contextlib.suppress(_ctypes.COMError):
            self.window.child_window(auto_id="txtTotApproved").set_text(str(ledger_allowed))

    def update_total_paid_header(self, ledger_payment: str) -> None:
        """Update Paid Amount Header in the Batch Header Window.

        Args:
            ledger_payment (str): The paid amount to be updated in the batch header.
        """
        self.logger.debug(f"Updating Paid Amount Header in the Batch Header Window to {ledger_payment}.")
        with contextlib.suppress(_ctypes.COMError):
            self.window.child_window(auto_id="txtTotPaid").set_text(str(ledger_payment))

    def set_user_in_secured_to(self, username: str) -> None:
        """Update combobox for user in the Batch Maintenace window."""
        self.logger.debug(f"Setting user in Secured To to {username}.")
        cbo_secured = self.window.child_window(auto_id="cboSecured", control_type="Pane")
        combobox_user = cbo_secured.child_window(auto_id="mcbo", control_type="ComboBox", found_index=0)
        self.desktop_app.select_value_from_combobox(combobox_user, username)

    def set_doc_mgmt(self, document_name: str) -> None:
        """Update combobox for image batch in the Batch Maintenace window."""
        self.logger.debug(f"Setting doc_mgmt to {document_name}.")
        cbo_image_batch = self.window.child_window(auto_id="cboImageBatch", control_type="Pane")
        combobox_image_batch = cbo_image_batch.child_window(auto_id="mcbo", control_type="ComboBox", found_index=0)
        self.select_combobox_doc_mgmt(combobox_image_batch, document_name)

    def select_combobox_doc_mgmt(
        self, combobox: ComboBoxWrapper, value: str, use_type_keys: bool = False, click_combobox: bool = True
    ) -> None:
        """Selects a value from a ComboBox by simulating a click and searching for the matching option.

        Args:
            combobox (ComboBoxWrapper): The ComboBox control in which to search for the value.
            value (str): The value to search for and select within the ComboBox.
            use_type_keys (bool, optional): If True, uses type_keys instead of set_text. Defaults to False.
            click_combobox (bool, optional): If True, clicks the combobox before selecting the value. Defaults to True.
        """
        current_value = combobox.legacy_properties().get("Value", "")
        if current_value:
            return
        if click_combobox:
            self.desktop_app.click_on_combobox(combobox)
        option_to_select = self.desktop_app.get_option_to_select_from_combobox(combobox, value)
        if option_to_select:
            self.desktop_app.select_option_in_combobox(combobox, use_type_keys, option_to_select)
        else:
            raise ValueError(f"Value '{value}' not found in ComboBox.")

    def update_default_tracking(self, text: str) -> None:
        """Updates the txtDefTracking field in the AMBatches window with the trn.

        Args:
            text (str): The text to be entered in the txtDefTracking field.
        """
        self.desktop_app.wait_until_element_visible(control_type="Window", auto_id="AMBatches")
        with contextlib.suppress(_ctypes.COMError):
            self.window.child_window(title="txtDefTracking", control_type="Edit").set_text(text)

    def update_batch_desc(self, text: str) -> None:
        """Updates the Batch Desc field.

        Args:
            text (str): new batch description text.
        """
        self.logger.debug(f"Updating Batch description in the Batch Header Window to {text}.")
        with contextlib.suppress(_ctypes.COMError):
            self.window.child_window(auto_id="txtDesc").set_text(text)

    def update_default_date(self, date_to_update: str) -> None:
        """Updates the default Date Field.

        Args:
            date_to_update (str): Date object to update field.
        """
        with contextlib.suppress(_ctypes.COMError):
            self.window.child_window(auto_id="txtDefDate").child_window(auto_id="mtxtDate").click_input()
            self.window.child_window(auto_id="txtDefDate").child_window(auto_id="mtxtDate").type_keys(date_to_update)

    def update_process_date(self, date_to_update: str) -> None:
        """Updates the Process Date Field.

        Args:
            date_to_update (str): Date object to update field.
        """
        with contextlib.suppress(_ctypes.COMError):
            self.window.child_window(auto_id="txtProcessDate").child_window(auto_id="mtxtDate").set_text(date_to_update)

    def update_total_adjustment(self, total_adjustment: str) -> None:
        """Update the total adjustment.

        Args:
            total_adjustment (str): total adjustment
        """
        try:
            self.desktop_app.wait_until_element_visible(control_type="Window", auto_id="txtTotAdjusted", retries=1)
        except PywinTimeout:
            self.logger.warning("Adjusted Header is not visible.")
            return

        self.logger.debug(f"Updating Adjusted Header in the Batch Header Window to {total_adjustment}.")
        with contextlib.suppress(_ctypes.COMError):
            self.window.child_window(auto_id="txtTotAdjusted").set_text(total_adjustment)



"""NextGenCore Module."""

import time
from typing import Optional
from retry import retry
from t_desktop.config import IS_WINDOWS_OS
from logging import Logger
import _ctypes
import contextlib

if IS_WINDOWS_OS:
    from pywinauto import keyboard
    from pywinauto.findwindows import ElementNotFoundError

from t_desktop.utils.capture_screenshot import capture_screenshot
from t_desktop.decorators import retry_if_pywin_error
from t_desktop.decorators import capture_screenshot_if_pywin_error
from t_desktop.exceptions import AppCrashException
from t_ocr import Textract
from t_nextgen.exceptions import WrongDatabaseSelectedError, MultipleSessionsError
from t_nextgen.config import Config
from t_nextgen.ng_app_manager import NGAppManager
from t_nextgen.ng_pm.windows.batch_maitenance import BatchMaintenanceWindow
from t_nextgen.ng_pm.windows.batch_posting import BatchPostingWindow
from t_nextgen.ng_pm.windows.cob import COBWindow
from t_nextgen.ng_pm.windows.encounter_insurance import EncounterInsuranceWindow
from t_nextgen.ng_pm.windows.import_file import ImportFileWindow
from t_nextgen.ng_pm.windows.payment_entry import PaymentEntryWindow
from t_nextgen.ng_pm.windows.transaction_ledger import TransactionLedgerWindow
from t_nextgen.utils.logger import logger as lib_logger


class NextGenPMCore:
    """NextGenPMCore Class.

    This class is used to access all the windows of NextGen PM.
    """

    def __init__(
        self,
        username: Optional[str] = "",
        password: Optional[str] = "",
        textract: Optional[Textract] = None,
        app_path: str = Config.DIRECTORIES.ENTERPRISE_PM,
        logger: Optional[Logger] = None,
    ):
        """This method is used to initialize the NextGenPMCore class.

        Args:
            username (Optional[str]): username of the user.
            password (Optional[str]): password of the user.
            textract (Optional[Textract], optional): Textract object. Defaults to None.
            app_path (str, optional): path of nextgen PM application. Defaults to Config.DIRECTORIES.ENTERPRISE_PM.
            logger (Optional[Logger], optional): logger object. Defaults to None.

        """
        self.username = username
        self.password = password
        self.app_path = app_path
        self.textract = textract
        self._batch_posting_window = None
        self._batch_maintenance_window = None
        self._cob_window = None
        self._encounter_insurance_window = None
        self._import_file_window = None
        self._payment_entry_window = None
        self._transaction_legder_window = None
        self.logger = logger if logger is not None else lib_logger
        self.desktop_app = NGAppManager(app_path, self.logger)

    @property
    def batch_posting_window(self):
        """This property is used to access BatchPostingWindow class."""
        if self._batch_posting_window is None:
            self._batch_posting_window = BatchPostingWindow(self.app_path, self.logger)
        return self._batch_posting_window

    @property
    def batch_maintenance_window(self):
        """This property is used to access BatchMaintenanceWindow class."""
        if self._batch_maintenance_window is None:
            self._batch_maintenance_window = BatchMaintenanceWindow(self.app_path, self.logger)
        return self._batch_maintenance_window

    @property
    def cob_window(self):
        """This property is used to access COBWindow class."""
        if self._cob_window is None:
            self._cob_window = COBWindow(self.app_path, self.logger)
        return self._cob_window

    @property
    def encounter_insurance_window(self):
        """This property is used to access EncounterInsuranteWindow class."""
        if self._encounter_insurance_window is None:
            self._encounter_insurance_window = EncounterInsuranceWindow(self.app_path, self.logger)
        return self._encounter_insurance_window

    @property
    def import_file_window(self):
        """This property is used to access ImportFileWindow class."""
        if self._import_file_window is None:
            self._import_file_window = ImportFileWindow(self.app_path, self.logger)
        return self._import_file_window

    @property
    def payment_entry_window(self):
        """This property is used to access PaymentEntryWindow class."""
        if self._payment_entry_window is None:
            self._payment_entry_window = PaymentEntryWindow(self.app_path, self.logger, self.textract)
        return self._payment_entry_window

    @property
    def transaction_ledger_window(self):
        """This property is used to access TransactionLedgerWiundow class."""
        if self._transaction_legder_window is None:
            self._transaction_legder_window = TransactionLedgerWindow(self.app_path, self.logger)
        return self._transaction_legder_window

    @retry_if_pywin_error(retries=2, delay=5)
    @capture_screenshot_if_pywin_error()
    def login(
        self,
        practice: Optional[str] = None,
        database: str = Config.DATABASES.TEST,
        app_folder: str = Config.DIRECTORIES.BASE_DIR,
    ) -> None:
        """Login to the app."""
        if process := self.desktop_app.get_app_session_if_running(self.app_path):
            self.logger.debug(f"Process {process} is already running.")
            capture_screenshot("app_session_is_running.png")
            self.close_session(process)  # Close a NextGen Session if it exists
        try:
            self.desktop_app.start_app(app_path=self.app_path, app_folder=app_folder)
            self.desktop_app.wait_until_element_visible(auto_id="txtUserName", control_type="Edit")
            self._check_database_name(db_name_flag=database)
            if not self.desktop_app.dialog.child_window(title="Windows Integrated", control_type="ComboBox").exists():
                self.desktop_app.set_input_text(auto_id="txtUserName", text=self.username)
                self.desktop_app.set_input_text(auto_id="txtPassword", text=self.password)
            if practice:
                self.desktop_app.select_dropdown_item("cboPractice", practice)
            self.desktop_app.invoke_button(auto_id="btnLogon")
            self.desktop_app.app = self.desktop_app.connect_to_app(title="MainForm", timeout=20)
            self.logger.info("Logged into NextGen")
        except Exception as e:
            try:
                if app := self.desktop_app.connect_to_app(title="MainForm", timeout=20):
                    self.logger.info("Logged into NextGen")
                    self.desktop_app.app = app
                    return
            except Exception as inner_e:
                self.handle_multiple_sessions_popup()
                self.desktop_app.close_login_window(self.app_path)
                raise inner_e
            self.logger.error(f"Login failed: {e}")
            raise e

    @retry_if_pywin_error(retries=3, delay=5, exceptions=(ElementNotFoundError))  # type: ignore[misc]
    def _check_database_name(self, db_name_flag: str = Config.DATABASES.PROD) -> None:
        """Validates if the name received is part of the database selected.

        Args:
            db_name_flag (str): The database name to be validated.

        Returns:
            Raises WrongDatabaseSelected: if the database doesn't match.
        """
        database_name = self.desktop_app.dialog.child_window(auto_id="lblDatabase").window_text()
        self.logger.info(f"App is connected to '{database_name}' database")
        if db_name_flag.lower() not in database_name.lower():
            raise WrongDatabaseSelectedError(f"'{database_name}' selected, wrong database selected.")

    def handle_multiple_sessions_popup(self) -> None:
        """Close multiple sessions pop-up that comes immediately after login.

        Raises:
            MultipleSessionsError: if the modal exists
        """
        if self.desktop_app.dialog.child_window(title="Multiple station logon error", control_type="Window").exists(
            timeout=3, retry_interval=1
        ):
            self.logger.warning("Multiple sessions detected. Closing the popup.")
            if next_gen_process := self.desktop_app.get_app_session_if_running(self.app_path):
                self.close_session(next_gen_process)
            raise MultipleSessionsError("Process failed with multiple sessions issue")

    def select_practice_from_app(self, practice: str):
        """Select practice from the Health Practice menu.

        Args:
            practice (str): The practice to select.
        """
        self.click_practice_from_bar()
        self.input_check_practice(practice)

    def check_if_practice_was_set_correctly(self, practice: str):
        """Check if the practice was set correctly.

        Args:
            practice (str): The practice to check.
        """
        pane = self.desktop_app.dialog.child_window(title="cboPractice", control_type="Pane").wait("visible", 5)
        practice_selected = pane.descendants(control_type="Edit")[0].get_value()
        if practice_selected != practice:
            raise Exception(f"Practice was not set correctly. Expected: {practice}, Actual: {practice}")

    @retry(tries=3, delay=3)
    def click_practice_from_bar(self):
        """Clicks the Health Practice button from the menu bar."""
        practice_icon = self.desktop_app.dialog.child_window(title="cmdToolPractice", control_type="Button")
        with contextlib.suppress(_ctypes.COMError):
            practice_icon.set_focus()
        with contextlib.suppress(_ctypes.COMError):
            practice_icon.click_input(button="left")
        self.desktop_app.dialog.child_window(auto_id="cboPractice", control_type="Pane").wait("visible", 5)

    @retry(tries=3, delay=3)
    def input_check_practice(self, practice: str):
        """Input check practice information and click OK.

        Args:
            practice (str): the name of the practice to be selected
        """
        pane = self.desktop_app.dialog.child_window(auto_id="cboPractice")
        combo_box = pane.child_window(auto_id="mcbo", control_type="ComboBox")
        with contextlib.suppress(_ctypes.COMError):
            combo_box.child_window(control_type="Edit", found_index=0).set_text(practice)
        self.check_if_practice_was_set_correctly(practice)
        with contextlib.suppress(_ctypes.COMError):
            self.desktop_app.dialog.child_window(auto_id="cmdOK").click()

    @capture_screenshot_if_pywin_error()
    def close_session(self, process: str, practice: str = None) -> None:
        """Close the session using keyboard shortcuts.

        Args:
            process (str): The name of the process to kill if necessary.

        Returns:
            None
        """
        self.logger.info("Running close_session method.")
        self.desktop_app.connect_to_main_form()
        self.desktop_app.close_modal(buttons_to_try=["NO", "OK"])
        self.desktop_app.clear_out_windows_function(practice=practice)
        self.desktop_app.open_file_menu()
        logout_key = "l"
        keyboard.send_keys(logout_key)

        # This kind of .exe needs to kill the app after logout

        self.desktop_app.close_modal(buttons_to_try=["YES"])
        time.sleep(20)
        try:
            self.desktop_app.wait_to_disappear(auto_id="cmdToolApptlist", control_type="Button", max_attempts=12)
        except AssertionError:
            raise AppCrashException("Session not closed successfully")
        self.logger.info("Session closed, killing the process for EnterprisePM.exe")
        time.sleep(5)
        self.desktop_app.kill_app(process=process, app_path=self.app_path)



"""PaymentEntry module."""
import contextlib
import _ctypes
from datetime import datetime
from logging import Logger
import os
from pathlib import Path
import re
from typing import Optional
from typing import Any

from decimal import Decimal
from t_desktop.config import IS_WINDOWS_OS

if IS_WINDOWS_OS:
    from pywinauto.application import WindowSpecification
    from pywinauto.controls.uia_controls import ListItemWrapper, EditWrapper
    from pywinauto.findwindows import ElementNotFoundError
    from pywinauto.keyboard import send_keys
    from pywinauto.timings import wait_until
from t_ocr import Textract
from t_ocr.pages import TextractPage
from PIL import Image, ImageFilter

from t_nextgen.nextgen_window import NextGenWindow
from t_nextgen.utils.convert_to_decimal import convert_to_decimal
from t_nextgen.utils.opencv import OpenCV
from t_nextgen.exceptions import (
    AdjFieldNotFoundException,
    BalanceFieldNotFoundException,
    Ins1FieldNotFoundException,
    Ins2FieldNotFoundException,
    Ins3FieldNotFoundException,
    PayFieldNotFoundException,
    ServiceLineMissingRequiredFieldError,
    TransactionNumberNotFoundError,
    DeductFieldNotFoundException,
    FieldNotFoundException,
)


class PaymentEntryWindow(NextGenWindow):
    """PaymentEntry class with methods to interact with payment entry window."""

    def __init__(self, app_path: str, logger: Logger, textract: Optional[Textract]):
        """Initialize the PaymentEntryWindow class.

        Args:
            app_path (str): The application path.
            textract (Optional[Textract]): The Textract object.
            logger (Logger): The logger object.

        """
        self.textract = textract
        super().__init__(app_path, logger)

    @property
    def window(self) -> WindowSpecification:
        """Return the PaymentEntry window element."""
        return self.desktop_app.dialog.child_window(title="PaymentEntry", control_type="Window")

    def set_all_service_lines_to_appeal(self) -> None:
        """This method sets all the service lines to appeal."""
        # Look at set_encounter_to_appeal_function
        self.logger.info("Setting Encounter to Appeal...")
        self.click_open_button()
        rows = self.get_service_rows()
        for row in rows:
            if self.get_pay_field(row):
                self.select_status_in_dropdown(status="Appeal", row=row, status_key="A")

    def get_pay_field(self, row: ListItemWrapper) -> str:
        """This method gets the pay field text in a specific row.

        Args:
            row (ListItemWrapper): list item wrapper with the row
        """
        items = row.descendants(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "pay" in window_text.lower():
                return item.get_value()
        raise PayFieldNotFoundException("Pay field not found in the row")

    def get_cpt4_field(self, row: ListItemWrapper) -> str:
        """This method gets the cpt 4 field text in a specific row."""
        items = row.descendants(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "cpt4" in window_text.lower():
                element = row.descendants(title=window_text, control_type="Edit")[0]
                return element.get_value()
        raise FieldNotFoundException("cpt 4 field not found in the row")

    def get_qty_charge_field(self, row: ListItemWrapper) -> str:
        """This method gets the qty/charge text in a specific row."""
        items = row.descendants(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "qty/charge" in window_text.lower():
                return item.get_value()
        raise FieldNotFoundException("qty/charge not found in the row")

    def get_date_field(self, row: ListItemWrapper) -> str:
        """This method gets the Date field text in a specific row."""
        items = row.descendants(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "date" in window_text.lower():
                return item.get_value()
        raise FieldNotFoundException("Date field not found in the row")

    def is_line_voided(
        self, row: ListItemWrapper, column_name: str = "Rendering", image_path: Path = Path("C:\\Temp\\")
    ) -> bool:
        """Check if the line is voided.

        Args:
            row: The row to be checked
            column_name: The column name to be checked (optional - default: "Rendering")

        Returns:
            bool: True if the line is voided, False otherwise
        """
        image_path.mkdir(parents=True, exist_ok=True)
        element = row.descendants(title=column_name)[0]
        is_line_voided = self.is_text_strikethrough(element, image_path)
        if is_line_voided:
            self.logger.debug("The line is voided")
        return is_line_voided

    @staticmethod
    def is_text_strikethrough(element: Any, image_path: Path) -> bool:
        """Check if the text in the element is strikethrough.

        Args:
            element: The element to be checked
        Returns:
            bool: True if the text is strikethrough, False otherwise
            image_path: Path to save the image
        """
        # element name with the hour min and sec
        element_name_with_date = f"{element.window_text()}_{datetime.now().strftime('%H_%M_%S_%f')}.png"
        element_image_path = os.path.join(image_path, element_name_with_date)
        image_element = element.capture_as_image()
        # Convert the image to grayscale
        grayscale_image = image_element.convert("L")
        # Sharpen the grayscale image to enhance the edges of the text
        sharpened_image = grayscale_image.filter(ImageFilter.SHARPEN)
        # Convert the sharpened image back to RGB
        sharpened_rgb_image = sharpened_image.convert("RGB")
        sharpened_rgb_image.save(element_image_path)
        is_strikethrough_present = OpenCV.check_strikethrough_in_text_image(element_image_path)
        os.remove(element_image_path)
        return is_strikethrough_present

    def click_open_button(self) -> None:
        """This method clicks on the open button."""
        self.logger.debug("Clicking on the open button.")
        open_button = self.desktop_app.dialog.child_window(title="_cmdAction_1", control_type="Button")
        if open_button.is_enabled():
            with contextlib.suppress(_ctypes.COMError):
                open_button.click()

    def get_service_rows(self) -> list[ListItemWrapper]:
        """This method gets the service rows from the current payment entry window.

        Returns:
            rows (list): list of service rows
        """
        pane = self.desktop_app.dialog.child_window(title="lstPayLedger", control_type="Pane")
        rows = pane.wrapper_object().descendants(control_type="DataItem")
        self.logger.debug(f"Number of rows in Payment Entry: {len(rows)}")
        return rows

    def insert_text_in_ln_item_rsns(self, text: str, row: ListItemWrapper) -> None:
        """This method inserts text in the Ln Item Rsns field.

        Args:
            text (str): text to be inserted
            row (ListItemWrapper): row to select. Will use the first row if not provided
        """
        self.logger.debug(f"Inserting text in Ln Item Rsns: {text}")
        if self.is_payer_a_patient():
            self.logger.info("The payer is Patient. Ln Item Rsns field is disabled")
            return

        self.remove_selected_items_from_ln_combobox(row)
        self.set_text_in_specific_row(row, "Ln Item Rsns", text, is_combobox=True)
        return row

    def remove_selected_items_from_status_combobox(self, row: ListItemWrapper) -> None:
        """Removes the selected items from the Status combobox.

        Args:
            row (ListItemWrapper): row to select
        """
        self.logger.debug("Removing selected items from Status combobox")
        row = self.revalidate_service_row(row)
        _input = row.descendants(title="Status", control_type="Edit")[0]

        if _input.get_value() == "":
            return
        with contextlib.suppress(_ctypes.COMError):
            _input.click_input(double=True)
        pane = self.window.child_window(title="_cboInput_0", control_type="Pane")
        combo_box = pane.child_window(control_type="ComboBox")
        combo_box.wait("ready", 2)
        self.desktop_app.dialog.child_window(
            auto_id="ColScrollRegion: 0, RowScrollRegion: 0", found_index=0
        ).right_click_input()
        send_keys("D")
        send_keys("{ESC}{TAB}")

    def is_payer_a_patient(self) -> bool:
        """This method checks if the payer is Patient.

        Returns:
            bool: True if payer is Patient, False otherwise
        """
        vse_container = self.window.child_window(auto_id="vseContainer", control_type="Text")
        cbo_input = vse_container.child_window(auto_id="_cboInput_104", control_type="Pane")
        mcbo = cbo_input.child_window(auto_id="mcbo", control_type="ComboBox")
        payer = mcbo.child_window(control_type="Edit", found_index=0)
        return payer.get_value().lower() == "patient"

    def select_status_in_dropdown(
        self, status: str = "Appeal", row: ListItemWrapper | None = None, status_key: str = ""
    ) -> None:
        """This method selects the specified status in the dropdown.

        Args:
            status (str): status to be selected (default is "Appeal")
            row (ListItemWrapper): row to select. Will use the first row if not provided
            status_key (str): Shortcut key for setting status (default sets to "Appeal" using "A"). Use
            only if first character of status is unique
        """
        self.logger.debug(f"Selecting status in dropdown: {status}")
        self.click_open_button()
        row = self.revalidate_service_row(row)
        status_field = self.get_status_field(row)
        if status_field.get_value() == status:
            self.logger.debug(f"Status was already set. Value: {status}")
            return
        pane = self.open_status_dropdown_in_service_row(status_field)
        self.select_status_in_service_row(pane, status, status_key)

    def select_status_in_service_row(self, pane: WindowSpecification, status: str, status_key: str = "") -> None:
        """Set status in the status column of current service row.

        Args:
            pane (WindowSpecification): Represents the pane element of the status field
            status (str): Status we want to set
            status_key (str, optional): Shortcut Key for setting status using first character of status. Use
            only if first character of status is unique
        """
        if status_key:
            send_keys(status_key)
            send_keys("{ENTER}")
        else:
            combo_box = pane.child_window(control_type="ComboBox")
            with contextlib.suppress(_ctypes.COMError):
                combo_box.child_window(auto_id="mcbo_EmbeddableTextBox", control_type="Edit").set_text(status)
                pane.click_input()

    def open_status_dropdown_in_service_row(self, status_field: EditWrapper) -> EditWrapper:
        """Open the status dropdown for the provided status field.

        Args:
            status_field (EditWrapper): the status field whose dropdown is to be opened

        Returns:
            WindowSpecification: Represents the pane element of the status field
        """
        self.logger.debug("Opening status dropdown in service row")
        with contextlib.suppress(_ctypes.COMError):
            status_field.click_input()
        pane = self.desktop_app.dialog.child_window(title="_cboInput_0", control_type="Pane")
        try:
            with contextlib.suppress(_ctypes.COMError):
                pane.click_input()
        except ElementNotFoundError:
            with contextlib.suppress(_ctypes.COMError):
                status_field.click_input()
            pane = self.desktop_app.dialog.child_window(title="_cboInput_0", control_type="Pane")
            with contextlib.suppress(_ctypes.COMError):
                pane.click_input()
        return pane

    def get_status_field(self, row: ListItemWrapper) -> EditWrapper:
        """Get status field from service row.

        Args:
            row (ListItemWrapper): The service row from which to fetch status field

        Returns:
            EditWrapper: the status field
        """
        status_field = row.descendants(title="Status", control_type="Edit")[0]
        return status_field

    def select_item_in_line_dropdown(self, row: ListItemWrapper, column_name: str, item: str) -> None:
        """This method selects the specified item in the dropdown.

        Args:
            row (ListItemWrapper): row to use
            column_name (str): column name
            item (str): item to be selected
        """
        self.logger.debug(f"Selecting item in dropdown: {item}")
        row = self._get_row_by_auto_id(row.automation_id())
        input = row.descendants(title=column_name, control_type="Edit")[0]
        if input.get_value() == str(item):
            return
        self.desktop_app.mouse_click_element(input, "left")
        self.desktop_app.mouse_click_element(input, "left")
        pane = self.desktop_app.dialog.child_window(title="_cboInput_0", control_type="Pane")
        combo_box = pane.child_window(control_type="ComboBox")
        combo_box.wait("ready", 2)
        edit_text = combo_box.child_window(control_type="Edit", found_index=0)
        with contextlib.suppress(_ctypes.COMError):
            edit_text.set_text(item)
        edit_text.type_keys("{ENTER}")
        combo_box.wait("ready", 2)
        input.set_focus()  # focus again in input to remove the dropdown from screen

    def _get_row_by_auto_id(self, auto_id: str) -> ListItemWrapper:
        """This method gets the row by auto_id from the current payment entry window.

        Args:
            auto_id (str): auto_id of the row
        Returns:
            row (ListItemWrapper): row object
        """
        pane = self.desktop_app.dialog.child_window(title="lstPayLedger", control_type="Pane")
        row = pane.child_window(control_type="DataItem", auto_id=auto_id)
        return row.wrapper_object()

    def set_transaction_notes_text(self, notes: str) -> None:
        """This method sets the transaction notes text.

        Args:
            notes (str): transaction notes text
        """
        self.logger.debug(f"Setting transaction notes text: {notes}")
        transaction_note_input = self.window.child_window(auto_id="txtTransactionNotes", control_type="Edit")
        with contextlib.suppress(_ctypes.COMError):
            transaction_note_input.set_edit_text(notes)

    def click_recalc_button(self, click_recalc_button: bool = True) -> None:
        """Clicks the recalc button."""
        if click_recalc_button:
            self.logger.debug("Clicking recalc button in Payment Entry Window.")
            self.desktop_app.click_button_by_element_name(element_name="_cmdAction_6")
        else:
            self.logger.warning("Skipping recalc button click to prevent changes.")

    def click_save_button(self, click_save_button: bool = True) -> None:
        """Clicks the save button."""
        if click_save_button:
            self.logger.debug("Clicking save button on Payment Entry Window.")
            self.desktop_app.click_button_by_element_name(element_name="_cmdAction_2")
            self.verify_and_handle_modal("Please specify a status for each line item payments and adjustment.")
        else:
            self.logger.warning("Clicking cancel in payment entry to prevent changes.")
            self.click_cancel()

    def click_cancel(self) -> None:
        """Clicks the cancel button on the payment entry window."""
        self.logger.debug("Clicking cancel button on Payment Entry Window.")
        self.desktop_app.click_button_by_element_name(element_name="_cmdAction_3")

    def verify_and_handle_modal(self, expected_text: str) -> None:
        """Verifies a modal's text and raises an exception if it matches the expected text.

        Args:
            expected_text (str): The expected text to verify in the modal.
        """
        success, modal_text = self.desktop_app.close_modal()

        if success and modal_text.upper() == expected_text.upper():
            message = f"'{expected_text}' found in the modal. Raising exception."
            self.click_cancel()
            raise ServiceLineMissingRequiredFieldError(message)

    def set_text_in_specific_row_to_adj_field(self, row: ListItemWrapper, text: str | float) -> None:
        """This method sets the text in a specific row to the adj field.

        The adj column can have different names on it. All of them have pay in the text.

        Args:
            row (ListItemWrapper): list item wrapper with the row
            text (str): text to be inserted
        """
        self.logger.debug(f"Setting text in Adj field: {text}")
        items = row.descendants(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "adj" in window_text.lower():
                self.set_text_in_specific_row(row, window_text, text)
                return
        raise AdjFieldNotFoundException("Adj field not found in the row")

    def set_text_in_specific_row(
        self, row: ListItemWrapper, element_name: str, text: str | float, is_combobox: bool = False
    ) -> None:
        """This method sets the text in a specific row.

        Args:
            row (ListItemWrapper): list item wrapper with the row
            element_name (str): element name
            text (str): text to be inserted
            is_combobox (bool): True if the element is a combobox, False otherwise
        """
        row = self.revalidate_service_row(row)
        element = row.descendants(title=element_name, control_type="Edit")[0]
        if element.get_value() == str(text):
            return
        self.desktop_app.mouse_double_click_element(element)
        with contextlib.suppress(_ctypes.COMError):
            wait_until(timeout=7, retry_interval=0.7, func=element.is_visible)
        self.desktop_app.mouse_double_click_element(element)
        with contextlib.suppress(_ctypes.COMError):
            wait_until(timeout=10, retry_interval=1, func=element.is_editable)
        with contextlib.suppress(_ctypes.COMError):
            element.set_text(text)
        self.desktop_app.mouse_double_click_element(element)

    def revalidate_service_row(self, row: ListItemWrapper, retries: int = 2) -> ListItemWrapper:
        """This method revalidates the service row in case the state of the row is not as expected.

        (i.e. the row is not visible or the coordinates are (0, 0))

        Args:
            row (ListItemWrapper): row to revalidate
            retries (int): number of retries
        Returns:
            row (ListItemWrapper): revalidated row
        """
        row_coords = self.desktop_app.get_element_coordinates(row)
        if row_coords == (0, 0) and retries > 0:
            refreshed_row = self._get_row_by_auto_id(row.automation_id())
            return self.revalidate_service_row(refreshed_row, retries=retries - 1)
        return row

    def get_transaction_notes_text(self) -> str:
        """This method gets the transaction notes text.

        Returns:
            str: transaction notes text
        """
        transaction_note_input = self.desktop_app.dialog.child_window(
            auto_id="txtTransactionNotes", control_type="Edit"
        )
        notes_value = transaction_note_input.get_value()
        return notes_value

    def get_pay_value_from_specific_row(self, row: ListItemWrapper) -> str:
        """This method gets a value from a specific row.

        Args:
            row (ListItemWrapper): list item wrapper with the row
        """
        items = row.descendants(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "pay" in window_text.lower():
                element = row.descendants(title=window_text, control_type="Edit")[0]
                return element.get_value()
        raise PayFieldNotFoundException("Pay Field not found")

    def get_adj_value_from_specific_row(self, row: ListItemWrapper) -> str:
        """Get the value from the "Adj" field in the row.

        Args
            matching_row (ListItemWrapper): if this argument is given, the method will
            search for the adj in this line.

        Returns:
            str: The value in the "Adj" field.
        """
        items = row.descendants(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "adj" in window_text.lower():
                adj = row.children(title=window_text, control_type="Edit")[0]
                return adj.get_value()
        raise AdjFieldNotFoundException("Adj Field not found")

    def get_ins1_value_from_specific_row(self, row: ListItemWrapper) -> str:
        """Get the value from the "Ins 1" field in the row.

        Args:
            row (ListItemWrapper): Will search for the Ins 1 in this line.

        Returns:
            str: The value in the "Ins 1" field.
        """
        items = row.descendants(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "ins 1" in window_text.lower():
                ins1 = row.children(title=window_text, control_type="Edit")[0]
                return ins1.get_value()
        raise Ins1FieldNotFoundException("Ins 1 field not found")

    def get_ins2_value_from_specific_row(self, row: ListItemWrapper) -> str:
        """Get the value from the "Ins 2" field in the row.

        Args:
            row (ListItemWrapper): Will search for the Ins 2 in this line.

        Returns:
            str: The value in the "Ins 2" field.
        """
        items = row.descendants(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "ins 2" in window_text.lower():
                ins2 = row.children(title=window_text, control_type="Edit")[0]
                return ins2.get_value()
        raise Ins2FieldNotFoundException("Ins 2 field not found")

    def get_ins3_value_from_specific_row(self, row: ListItemWrapper) -> str:
        """Get the value from the "Ins 3" field in the row.

        Args:
            row (ListItemWrapper): Will search for the Ins 3 in this line.

        Returns:
            str: The value in the "Ins 3" field.
        """
        items = row.descendants(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "ins 3" in window_text.lower():
                ins3 = row.children(title=window_text, control_type="Edit")[0]
                return ins3.get_value()
        raise Ins3FieldNotFoundException("Ins 3 field not found")

    def get_balance_value_from_specific_row(self, row: ListItemWrapper) -> str:
        """Get the value from the "Balance" field in the row.

        Args:
            row (ListItemWrapper): Will search for the Balance in this line.

        Returns:
            str: The value in the "Balance" field.
        """
        items = row.descendants(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "balance" in window_text.lower():
                balance = row.children(title=window_text, control_type="Edit")[0]
                return balance.get_value()
        raise BalanceFieldNotFoundException("Balance field not found")

    def remove_selected_items_from_ln_combobox(self, row: ListItemWrapper) -> None:
        """Removes the selected items from the Ln Item Rsns combobox.

        Args:
            row (ListItemWrapper): row to select
        """
        row = self.revalidate_service_row(row)
        _input = row.descendants(title="Ln Item Rsns", control_type="Edit")[0]

        if _input.get_value() == "":
            return
        with contextlib.suppress(_ctypes.COMError):
            _input.click_input()
        pane = self.window.child_window(title="_cboInput_0", control_type="Pane")
        try:
            with contextlib.suppress(_ctypes.COMError):
                pane.click_input()
        except ElementNotFoundError:
            with contextlib.suppress(_ctypes.COMError):
                _input.click_input()
            pane = self.window.child_window(title="_cboInput_0", control_type="Pane")
            with contextlib.suppress(_ctypes.COMError):
                pane.click_input()
        with contextlib.suppress(_ctypes.COMError):
            self.desktop_app.app.window(title="MainForm").child_window(
                title="", control_type="Window"
            ).right_click_input()
        send_keys("D")
        send_keys("{ESC}{TAB}")

    def maximize_window(self) -> None:
        """Maximize the Payment Entry Window."""
        if not self.window.is_maximized():
            self.logger.debug("Maximizing Payment Entry Window.")
            self.window.maximize()

    def click_new_button(self) -> None:
        """Click New Button."""
        self.logger.debug("Clicking New Button.")
        self.desktop_app.click_button_by_element_name(element_name="_cmdAction_0")

    def enter_text_in_enc_clm_field(self, text: str) -> None:
        """This method inserts text in the Enc/Clm #.

        Args:
            text (str): text to be inserted
        """
        self.logger.debug(f"Inserting text in Enc/Clm #: {text}")
        with contextlib.suppress(_ctypes.COMError):
            container = self.desktop_app.dialog.child_window(auto_id="vseContainer", control_type="Text")
            container.set_focus()
            container.child_window(title="_txtActiveSearch_103", control_type="Edit").set_text(text)
            send_keys("{ENTER}")

    def enter_text_in_tracking_field(self, text: str) -> None:
        """This method inserts text in the Enc/Clm #.

        Args:
            text (str): text to be inserted
        """
        self.logger.debug(f"Inserting text in Tracking #: {text}")
        with contextlib.suppress(_ctypes.COMError):
            container = self.desktop_app.dialog.child_window(auto_id="vseContainer", control_type="Text")
            container.set_focus()
            container.child_window(title="_txtDataEntry_105", control_type="Edit").set_text(text)
            send_keys("{ENTER}")

    def click_close_nextgen_alerts(self) -> None:
        """Click close button on NextGen Alerts."""
        alert = self.desktop_app.dialog.child_window(title="NextGen Alerts", control_type="Window")
        if alert.exists(timeout=5, retry_interval=1):
            self.logger.debug("Clicking close on NextGen Alerts")
            alert.set_focus()
            with contextlib.suppress(_ctypes.COMError):
                alert.child_window(title="Close", control_type="Button").click()

    def click_yes_encounter_in_history_status_pop_up(self) -> None:
        """Click yes button in the encounter in history status pop-up."""
        if self.desktop_app.dialog.child_window(
            title="The encounter is in a history status.  Posting a transaction(s) against this encounter "
            "will reinstate the encounter to a billed status.  Do you want to continue?",
            control_type="Text",
        ).exists(timeout=1, retry_interval=1):
            self.logger.debug("Clicking Yes in Encounter in History Status pop-up")
            with contextlib.suppress(_ctypes.COMError):
                self.desktop_app.dialog.child_window(title="Yes", control_type="Button").click_input()

    def _capture_pages_data_as_image_in_window(self) -> Image:
        """Capture Pages Data as Image.

        Returns:
            Image: the captured image
        """
        container = self.desktop_app.dialog.child_window(auto_id="vseContainer", control_type="Text")
        custom_control = container.child_window(auto_id="_lblDisplayField_0", control_type="Text")
        return custom_control.capture_as_image()

    def get_pages_data(self, screenshot_destination: Path) -> str:
        """Extract Pages Dataw.

        Returns:
            str: the pages data as text
        """
        image = self._capture_pages_data_as_image_in_window()
        return self.get_data_from_image(image, screenshot_destination)

    def get_data_from_image(self, image: Image.Image, screenshot_destination: Path) -> str:
        """Extract text from image after preprocessing it.

        Args:
            image (Image.Image): The image from which to extract text.
            parsing_settings (PSM): The parsing settings.

        Returns:
            str: The extracted text.
        """
        if self.textract is None:
            raise ValueError("You must provide textract in Core Class initialization.")
        screenshot_name = f'temp_screenshot_{datetime.now().strftime("%Y%m%d_%H%M%S")}.png'
        screenshot_location = screenshot_destination / screenshot_name
        image.save(screenshot_location)

        # Perform OCR
        ocrd_image: TextractPage = self.textract.read_image_page(screenshot_location, cache_data=False)

        # Clean up the temporary image file
        screenshot_location.unlink()

        return ocrd_image.full_text

    def get_transaction_number(self, screenshot_destination: Path) -> Optional[str]:
        """Extracts and returns the transaction number from the page data using regex.

        Returns:
            Optional[str]: The transaction number if found, otherwise None.
        """
        self.maximize_window()
        self.desktop_app.wait_until_element_visible(auto_id="vseContainer", control_type="Text")
        page_data = self.get_pages_data(screenshot_destination)

        match = re.search(r"(\d+)\s*of\s*\d+", page_data)
        if match:
            txn_value = "00" + match.group(1)
            self.logger.debug(f"Transaction Number: {txn_value}")
            return txn_value

        raise TransactionNumberNotFoundError("Transaction number not found in the page data.")

    def click_shield_icon(self) -> None:
        """Click shield icon in pyament entry window."""
        self.logger.debug("Clicking shield icon in Payment Entry Window.")
        pnl_small_tool = self.window.child_window(title="pnlSmallTool", control_type="Text")
        with contextlib.suppress(_ctypes.COMError):
            pnl_small_tool.child_window(auto_id="_cmdSmallTool_03", control_type="Button").click_input()

    def select_payer_in_drop_down(self, payer: str) -> None:
        """This method selects the payer in the drop down.

        Args:
            payer (str): payer to be selected
        """
        self.logger.debug(f"Selecting payer in dropdown: {payer}")
        container = self.window.child_window(auto_id="vseContainer", control_type="Text")
        pane = container.child_window(title="_cboInput_104", control_type="Pane")
        combo_box = pane.child_window(auto_id="mcbo", control_type="ComboBox")
        pane_object = pane.wrapper_object()
        options = pane_object.descendants(control_type="ListItem")
        for option in options:
            value = option.window_text()
            if payer.lower() in value.lower():
                with contextlib.suppress(_ctypes.COMError):
                    combo_box.child_window(control_type="Edit", found_index=0).set_text(value)
                return

    def get_payers_from_dropdown(self) -> list:
        """Get all the Payer element from the dropdown.

        Returns:
            list: List of payers as ListItem
        """
        container = self.window.child_window(auto_id="vseContainer", control_type="Text")
        pane = container.child_window(title="_cboInput_104", control_type="Pane")
        pane_object = pane.wrapper_object()
        return pane_object.descendants(control_type="ListItem")

    def set_text_in_allowed(self, row: ListItemWrapper, allowed_amt: Decimal) -> None:
        """Update the Allowed field for a specific service line."""
        self.logger.debug(f"Setting Allowed value in specific row: {allowed_amt}")
        items = row.descendants(control_type="Edit")
        allowed_element_name = None
        for item in items:
            window_text = item.window_text()
            if "allowed" in window_text.lower():
                allowed_element_name = window_text
                self.set_text_in_specific_row(row, allowed_element_name, str(allowed_amt))
                self.skipping_payer_aarp_pop_up_exist()
                return

    def set_text_in_specific_row_to_pay_field(self, row: ListItemWrapper, text: str | float) -> None:
        """This method sets the text in a specific row to the pay field.

        Args:
            row (ListItemWrapper): list item wrapper with the row
            text (str): text to be inserted
        """
        self.logger.debug(f"Setting text in Pay field: {text}")
        with contextlib.suppress(_ctypes.COMError):
            str_text = str(convert_to_decimal(text))
            items = row.descendants(control_type="Edit")
            payment_element_name = None
            for item in items:
                window_text = item.window_text()
                if "pay" in window_text.lower():
                    payment_element_name = window_text
                    self.set_text_in_specific_row(row, payment_element_name, str_text)
                    self.skipping_payer_aarp_pop_up_exist()
                    return
            raise PayFieldNotFoundException("Pay field not found in the row")

    def skipping_payer_aarp_pop_up_exist(self) -> bool:
        """This method checks if the skipping payer AARP pop up exists. If it does, it clicks on the ok button.

        Returns:
            bool: True if the pop up exists, False otherwise
        """
        if self.desktop_app.dialog.child_window(auto_id="65535", control_type="Text").exists(
            timeout=3, retry_interval=1
        ):
            with contextlib.suppress(_ctypes.COMError):
                self.desktop_app.dialog.child_window(title="OK", control_type="Button").click()
            self.logger.debug("Clicked on OK button in Payer AARP pop up.")
            return True
        return False

    def get_next_gen_payer(self) -> str:
        """Get NextGen Payer on Payment Entry Window."""
        vse_container = self.window.child_window(auto_id="vseContainer", control_type="Text")
        cbo_input = vse_container.child_window(auto_id="_cboInput_104", control_type="Pane")
        mcbo = cbo_input.child_window(auto_id="mcbo", control_type="ComboBox", found_index=0)
        nextgen_payer = mcbo.child_window(auto_id="[Editor] Edit Area", control_type="Edit").get_value()
        match = re.search(r"^(.*?)\/", nextgen_payer)
        if match:
            return match.group(1)
        self.logger.debug(f"Payer: {nextgen_payer}")
        return nextgen_payer

    def get_deduct_value_from_specific_row(self, row: ListItemWrapper) -> str:
        """Get the value from the "Deduct" field in the row.

        Args:
            row (ListItemWrapper): Will search for the Balance in this line.

        Returns:
            str: The value in the "Deduct" field.
        """
        items = row.descendants(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "deduct" in window_text.lower():
                deduct = row.children(title=window_text, control_type="Edit")[0]
                self.logger.debug(f"Deduct value: {deduct.get_value()}")
                return deduct.get_value()
        raise DeductFieldNotFoundException("Deduct field not found")

    def set_text_in_deduct_field(self, row: ListItemWrapper, deduct_amt: Decimal) -> None:
        """Update the Deduct field for a specific service line."""
        self.logger.debug(f"Setting Deduct value in specific row: {deduct_amt}")
        items = row.descendants(control_type="Edit")
        for item in items:
            window_text = item.window_text()
            if "deduct" in window_text.lower():
                allowed_element_name = window_text
                self.set_text_in_specific_row(row, allowed_element_name, str(deduct_amt))
                self.skipping_payer_aarp_pop_up_exist()
                return

    def refresh_and_enter_encounter_num_and_trn(self, encounter_num: str, payment_trn: str) -> None:
        """This function closes and reopens the payment.

        Args:
            encounter_num (str): Encounter number
            payment_trn (str): Payment TRN
        """
        self.logger.debug("Refreshing the PaymentEntry window.")
        with contextlib.suppress(_ctypes.COMError):
            self.logger.debug("Clean the PaymentEntry window with ALT+L.")
            self.window.type_keys("%l")
            self.logger.debug(f"Entering encounter number: {encounter_num} in PaymentEntry window.")
            self.enter_text_in_enc_clm_field(encounter_num)
            self.click_close_nextgen_alerts()
            self.click_yes_encounter_in_history_status_pop_up()
            self.enter_text_in_tracking_field(payment_trn)
            self.logger.debug("Payment window cleaned and tracking number entered successfully.")

    def select_pay_code_in_drop_down(self, option_to_select: str):
        """This method selects the pay code in the drop down.

        Args:
            option_to_select (str): pay code to be selected
        """
        self.logger.debug(f"Selecting payer in dropdown: {option_to_select}")
        container = self.window.child_window(auto_id="vseContainer", control_type="Text")
        pane = container.child_window(title="_cboInput_112", control_type="Pane")
        combo_box = pane.child_window(auto_id="mcbo", control_type="ComboBox")
        pane_object = pane.wrapper_object()
        options = pane_object.descendants(control_type="ListItem")
        for option in options:
            value = option.window_text()
            if option_to_select.lower() in value.lower():
                with contextlib.suppress(_ctypes.COMError):
                    combo_box.child_window(control_type="Edit", found_index=0).set_text(value)
                return

    def select_adj_code_in_drop_down(self, option_to_select: str):
        """Select the adjustment code from the dropdown.

        Args:
            option_to_select (str): the adjustment code to be selected
        """
        container = self.window.child_window(auto_id="vseContainer", control_type="Text")
        pane = container.child_window(title="_cboInput_113", control_type="Pane")
        combo_box = pane.child_window(auto_id="mcbo", control_type="ComboBox")
        pane_object = pane.wrapper_object()
        options = pane_object.descendants(control_type="ListItem")
        for option in options:
            value = option.window_text()
            if option_to_select.lower() in value.lower():
                with contextlib.suppress(_ctypes.COMError):
                    combo_box.child_window(control_type="Edit", found_index=0).set_text(value)
                return

    def close_payment_entry_window(self):
        """Click close payment entry window."""
        title_bar = self.window.child_window(control_type="TitleBar")
        title_bar.child_window(title="Close", control_type="Button").click_input()

    def insert_control_number_in_resub(self, control_number: str) -> None:
        """This method inserts the control number in the resubmission field.

        Args:
            control_number (str): _description_
        """
        self.desktop_app.child_window(title="_txtDataEntry_116", control_type="Edit").set_text(control_number)

    def click_show_all_lines_check_box(self) -> None:
        """This method clicks on the show all lines check box."""
        check_box = self.desktop_app.child_window(title="chkShowAll", control_type="CheckBox")
        if check_box.get_toggle_state() == 0:
            try:
                check_box.toggle()
            except _ctypes.COMError:  # type: ignore[attr-defined]
                pass

    def check_bad_debt(self) -> bool:
        """This method checks if the encounter is pre-listed for bad debt.

        Returns:
            bool: True if the pop up exists, False otherwise
        """
        bad_debt_popup = self.desktop_app.dialog.child_window(title="NextGen", control_type="Window")
        bad_debt_text = self.desktop_app.child_window(
            title="This Encounter is pre-listed for bad debt or in bad debt.  You may not edit any information."
        )
        if bad_debt_text.exists():
            bad_debt_popup.child_window(title="OK", control_type="Button").click_input()
            balance_window = self.desktop_app.child_window(title="BalanceControl", control_type="Window").wait(
                "visible", 10
            )
            balance_window.child_window(title="cmdCancel", control_type="Button").click_input()
            return True
        return False

    def check_invalid_encounter_pop_up_exist(self) -> bool:
        """This method checks if the invalid encounter pop up exists.

        Returns:
            bool: True if the pop up exists, False otherwise
        """
        invalid_encounter_popup = self.desktop_app.dialog.child_window(title="NextGen", control_type="Window")
        invalid_encounter_text_1 = self.desktop_app.dialog.child_window(
            title="The Encounter number entered is invalid."
        )
        invalid_encounter_text_2 = self.desktop_app.dialog.child_window(title="Invalid encounter number!")
        if invalid_encounter_text_1.exists(timeout=3, retry_interval=1) or invalid_encounter_text_2.exists(
            timeout=1, retry_interval=1
        ):
            invalid_encounter_popup.child_window(title="OK", control_type="Button").click_input()
            return True
        return False



"""Module for Batch Posting Window."""
import re
import time
import contextlib
from decimal import Decimal

import _ctypes
from typing import Optional
import pyautogui
from retry import retry
from t_desktop.config import IS_WINDOWS_OS
from t_desktop.utils.capture_screenshot import capture_screenshot

if IS_WINDOWS_OS:
    from pywinauto.keyboard import send_keys
    from pywinauto.controls.uia_controls import ListItemWrapper
    from pywinauto.application import WindowSpecification
    from pywinauto.findwindows import ElementNotFoundError
from t_desktop.decorators import retry_if_pywin_error
from t_nextgen.nextgen_window import NextGenWindow
from t_nextgen.constants import CONST
from t_nextgen.exceptions import (
    NoMatchingBatchDescriptionException,
    NextGenDuplicateImportException,
    BatchFromBarNotFound,
)

from t_nextgen.ng_pm.windows.models.batch_posting_row import BatchPostingRow


class BatchPostingWindow(NextGenWindow):
    """Batch Posting Class with methods to interact with batch posting window."""

    @property
    def window(self):
        """Return the AMBatches window element."""
        self.desktop_app.dialog.child_window(auto_id="NGEPMBatchLookup", control_type="Window").wait("visible", 5)
        return self.desktop_app.dialog.child_window(auto_id="NGEPMBatchLookup", control_type="Window")

    @property
    def batch_posting_window(self):
        """Return the BatchPosting window element."""
        return self.desktop_app.dialog.child_window(title="BatchPosting", control_type="Window")

    @retry(tries=3, delay=2)
    def click_batch_icon_from_bar(self, practice_name: str) -> None:
        """Clicks the Batch Posting button from the menu bar."""
        try:
            self.logger.debug("Clicking on the batch icon from the bar.")
            batch_icon = self.desktop_app.dialog.child_window(auto_id="cmdToolPosting")
            self.desktop_app.mouse_click_element(batch_icon)
            self.desktop_app.safe_wait({"auto_id": "Data Area", "control_type": "Custom", "found_index": 0})
            self.maximize_batch_window()
        except _ctypes.COMError:
            pass
        except TimeoutError as e:
            self.desktop_app.deal_with_unhandled_exception_popup(f"NextGen - {practice_name}")
            raise e
        except ElementNotFoundError:
            raise BatchFromBarNotFound("Element with auto_id: cmdToolPosting, not found")

    def post_batch(self, row: ListItemWrapper):
        """Function for Posts the batch.

        Post a batch by right-clicking the given row and selecting the 'Post' option
        from the context menu, if the button is enabled.

        Args:
            row (ListItemWrapper): The UI element representing the batch row to be posted.
        """
        self.maximize_batch_window()
        row.click_input(button="right")
        button = self.desktop_app.dialog.child_window(title_re="Post", control_type="MenuItem")
        if button.is_enabled():
            button.click_input()
            self.desktop_app.close_modal(buttons_to_try=["OK"])
            return True
        else:
            self.logger.warning("Button disabled unable to post the batch.")
            return False

    def click_batch_from_window_menu_if_exist(self) -> bool:
        """This method clicks on the batch from the window menu.

        Returns:
            bool: True if the window menu is found, False otherwise
        """
        windows_menu = self.desktop_app.dialog.child_window(title="Window", control_type="MenuItem")
        windows_menu.set_focus()
        windows_menu.select()
        try:
            windows_menu.child_window(title="1 Batch Posting", control_type="MenuItem").select()
            self.logger.debug("Clicked on the batch from the window menu")
            return True
        except _ctypes.COMError:
            return True
        except ElementNotFoundError:
            return False

    def _deal_with_no_batch_found_popup(self) -> None | NoMatchingBatchDescriptionException:
        modal = self.desktop_app.dialog.child_window(title="NextGen", control_type="Window")
        if modal.exists(timeout=3, retry_interval=1):
            self.logger.debug("No batch found. Closing popup")
            send_keys("{ENTER}")
            send_keys("%C")
            raise NoMatchingBatchDescriptionException(CONST.BATCH_DESCRIPTION_NOT_FOUND)
        return None

    def _set_amount_field_in_batch_lookup(self, search_criteria_window: WindowSpecification, amount: Decimal) -> None:
        self.logger.debug(f"Setting amount field in batch lookup to {amount}")
        amount_field = search_criteria_window.child_window(auto_id="txtAmount", control_type="Edit")
        amount_field.set_focus()
        with contextlib.suppress(_ctypes.COMError):
            amount_field.set_text(amount)

    def find_batch(
        self, description: str, amount: Optional[Decimal] = None, fill_amount: bool = False
    ) -> ListItemWrapper:
        """Find a batch in the Batch Posting window.

        Args:
            description (str): The description of the batch to find.
            amount (Optional[Decimal]): The amount of the batch to find.
            fill_amount (Optional[bool]): Whether to fill the amount field before searching.
        """
        batch_row = self.run_batch_lookup(description, amount, fill_amount)
        return batch_row.row_element

    def run_batch_lookup(self, trn: str, amount: Optional[Decimal], fill_amount: bool = False) -> BatchPostingRow:
        """Performs a batch lookup based on the provided criteria.

        Args:
            trn (str): The batch description to search for.
            amount (Decimal): The amount associated with the batch, used to filter the results.
            fill_amount (bool): If True, fills in the amount field in the search interface before performing the search.
                                Default is False.

        Returns:
            BatchPositngRow: Returns BatchPostingRow object or raises an exception if row not found

        Raises:
            NoMatchingBatchDescriptionException: If no matching batch is found.
            NextGenDuplicateImportException: If more than one matching batch is found.
        """
        self.logger.info(f"Running batch lookup: trn={trn}, amount={amount}")
        self.click_menu_icon("a")
        search_criteria_window = self.window.child_window(auto_id="grpSearchCriteria")
        if fill_amount:
            self._set_amount_field_in_batch_lookup(search_criteria_window, amount)
        description_field = search_criteria_window.child_window(auto_id="txtDesc", control_type="Edit")
        description_field.set_focus()
        with contextlib.suppress(_ctypes.COMError):
            description_field.set_text(trn)
        send_keys("%F")

        self._deal_with_no_batch_found_popup()

        tree_view = self.window.child_window(auto_id="ColScrollRegion: 0, RowScrollRegion: 0")

        filtered_rows = self.get_batch_posting_rows(tree_view)
        if filtered_rows is None:
            send_keys("%C")
            self.logger.debug("No matching batch description found")
            raise NoMatchingBatchDescriptionException(CONST.BATCH_DESCRIPTION_NOT_FOUND)

        rows = self._filter_batch_posting_rows(filtered_rows, trn, amount)
        rows_count = len(rows)
        if rows_count == 1:
            row: BatchPostingRow = rows[0]
            if not row.visible:
                self._click_down_n_times(row.index)
                row.visible = True
            with contextlib.suppress(_ctypes.COMError):
                row.edit_wrapper.click_input()
            self.logger.info("Batch row found")
            return row
        elif rows_count == 0:
            send_keys("%C")
            raise NoMatchingBatchDescriptionException(CONST.BATCH_DESCRIPTION_NOT_FOUND)
        else:
            send_keys("%C")
            self.logger.debug("More than one batch with the same description was found")
            capture_screenshot(file_name=f"{trn}_duplicated_batch.png")
            raise NextGenDuplicateImportException("More than one batch with the same description was found")

    def click_menu_icon(self, key: str) -> None:
        """Clicks the menu icon button in the 'Batch Posting' window.

        Args:
            batch_window (WindowSpecification): The 'Batch Posting' window object.
            key (str): The key to send after clicking the button.
        """
        self.logger.debug("Clicking the menu icon button in the 'Batch Posting' window")
        self.batch_window = self._get_batch_window()
        cmd_drill_button = self.batch_window.child_window(title="cmdDrill", control_type="Button")
        with contextlib.suppress(_ctypes.COMError):
            cmd_drill_button.click_input()
        time.sleep(1)

        send_keys(key)

    @retry(tries=3, delay=1)
    def _get_batch_window(self) -> WindowSpecification:
        """Retrieves the 'Batch Posting' window.

        Returns:
            WindowSpecification: The 'Batch Posting' window object.
        """
        self.logger.debug("Getting the 'Batch Posting' window")
        self.desktop_app.wait_until_element_visible(control_type="Window", title="BatchPosting", timeout=15)
        self.desktop_app.wait_until_element_visible(control_type="Pane", title="lstListing", timeout=15)
        batch_window = self.desktop_app.dialog.child_window(title="BatchPosting", control_type="Window")
        return batch_window

    def get_batch_posting_rows(self, rows_list: ListItemWrapper) -> list[BatchPostingRow]:
        """The function retrieves about the Batch Posting rows.

        :param rows_list: The `rows_list` parameter is the object representing the tree view control in
        the user interface. It is used to access the elements and descendants of the tree view
        :return: a list of filtered rows consisting of the BatchPostingRow object
        """
        filtered_rows = []
        all_descendants = rows_list.descendants()

        index = 0
        for element in all_descendants:
            if "DataItem" in element.element_info.control_type:
                index += 1
                data_item_descendants = element.descendants(control_type="Edit")
                batch_posting_row = BatchPostingRow(
                    text=element.element_info.name,
                    visible=element.is_visible(),
                    index=index,
                    row_element=element,
                )
                for descendant in data_item_descendants:
                    if "Description" in descendant.element_info.name:
                        batch_posting_row.description = descendant.get_value()
                        batch_posting_row.edit_wrapper = descendant
                    if "Members" in descendant.element_info.name:
                        batch_posting_row.member_count = descendant.get_value()
                    if batch_posting_row.member_count and batch_posting_row.description:
                        break
                filtered_rows.append(batch_posting_row)
        return filtered_rows

    def _filter_batch_posting_rows(
        self, batch_posting_rows: list[BatchPostingRow], description: str, amt: Decimal
    ) -> list[BatchPostingRow]:
        """Filter Batch Posting rows based on description and amount text in BatchPostingRow.description.

        Args:
            dicts (list[dict]): A list of dictionaries to search through.
            description (str): The description to match within each dictionary.
            amt (Decimal): Amount from the payment object

        Returns:
            list[BatchPostingRow]: A list of BatchPostingRow where the description matches the given description.
        """
        result = []
        for row in batch_posting_rows:
            if description not in row.description:
                continue
            amt_extracted = self._extract_amt_from_description(row.description)
            if amt_extracted is None:
                self.logger.info("Unable to determine amount value from batch description")
            elif amt == Decimal(amt_extracted):
                result.append(row)
        return result

    def _extract_amt_from_description(self, description: str) -> None | str:
        """Extract amount from batch description in nextgen.

        Args:
            description (str): The description from which we want the amount.

        Returns
            (str): The amount extracted.
        """
        # Find the amount with optional commas and dollar sign
        match = re.search(r"\$\s?([\d,]+\.?\d*)", description)
        if match:
            # Extract the numeric part and replace commas
            amount_str = match.group(1).replace(",", "")
            self.logger.debug(f"Extracted amount from description: {amount_str}")
            return amount_str
        return None

    def _click_down_n_times(self, n: int, confidence_value: int = 3) -> None:
        """This method clicks the down key n times.

        Args:
            n (int): The number of times to click the down key.
            confidence_value (int, optional): confidence value. Defaults to 3.
        """
        self.logger.debug(f"Clicking down {n + confidence_value} times")
        pyautogui.press("home")
        for _ in range(n + confidence_value):
            pyautogui.press("down")

    def check_boxes_in_batch_posting(self, current_phase: str) -> None:
        """Toggle the state of three checkboxes in a batch posting dialog.

        Toggles the state of checkboxes with auto_ids "_chkSecure_0", "_chkSecure_1", and "_chkSecure_2"
        in the current dialog.

        Args:
            current_phase: the current phase of the work item
        """

        @retry(tries=3, delay=3)
        def toggle_checkbox(checkbox, untoggle: bool = False):
            """This method toggle the checkbox and prevents the useless COMError.

            This also wraps the necessity of using one Try-Except for each checkbox

            Args:
                checkbox (pywinauto.controls.uia_controls.CheckBoxWrapper): checkbox object
                untoggle (bool): untoggle the checkbox if True
            """
            with contextlib.suppress(_ctypes.COMError):
                toggle_value = 1 if untoggle else 0
                if checkbox.get_toggle_state() == toggle_value:
                    checkbox.toggle()

        checkbox_ids = ["_chkSecure_0", "_chkSecure_1", "_chkSecure_2"]
        checkboxes = []
        checkboxes.append(
            self.desktop_app.dialog.child_window(auto_id=checkbox_ids[0], control_type="CheckBox").wait("visible", 10)
        )
        for checkbox_id in checkbox_ids:
            checkboxes.append(self.desktop_app.dialog.child_window(auto_id=checkbox_id, control_type="CheckBox"))

        for checkbox in checkboxes:
            untoggle = False
            if checkbox == "_chkSecure_0" and current_phase == CONST.WORK_ITEM_PHASE.RETRY.value:
                untoggle = True
            toggle_checkbox(checkbox, untoggle)

    @retry(tries=3, delay=3)
    def set_focus_in_batch_window(self):
        """This method sets the focus in the batch window."""
        self.logger.debug("Setting focus in the batch window")
        window = self.desktop_app.dialog.child_window(title="BatchPosting", control_type="Window")
        pane = window.child_window(auto_id="lstListing")
        div = pane.child_window(auto_id="Data Area", control_type="Custom", found_index=0)
        tree_view = div.child_window(auto_id="ColScrollRegion: 0, RowScrollRegion: 0")
        self.desktop_app.mouse_click_element(tree_view)

    @retry(tries=3, delay=3)
    def click_ok_quick_batch(self) -> None:
        """The function clicks on the "OK" button in the Quick Batch dialog window."""
        self.logger.debug("Clicking on the OK button in the Quick Batch dialog window")
        ok_button = self.desktop_app.dialog.child_window(title="cmdOK", control_type="Button")
        with contextlib.suppress(_ctypes.COMError):
            ok_button.click()

    @staticmethod
    def click_quick_batch_from_batch_menu(quick_batch_position: int = 2) -> None:
        """The function clicks on the "Quick Batch" option in the batch menu."""
        for _ in range(quick_batch_position):
            send_keys("{DOWN}")
        send_keys("{ENTER}")

    @staticmethod
    def click_new_from_batch_menu(new_batch_item_position: int = 1) -> None:
        """The function clicks on the "New..." option in the batch menu.

        Args:
            new_batch_item_position (int): the position of the "New..." option in the batch menu
        """
        for _ in range(new_batch_item_position):
            send_keys("{DOWN}")
        send_keys("{RIGHT}")

    def maximize_batch_window(self) -> None:
        """This method maximizes the batch window."""
        try:
            if not self.batch_posting_window.was_maximized():
                self.logger.debug("Maximizing the batch window")
                self.batch_posting_window.maximize()
            time.sleep(2)
        except Exception as e:
            self.logger.warning(f"Could not maximize the batch window: {str(e)}")

    def get_batch_row(self, trn: str, amount: str) -> None | BatchPostingRow:
        """This method retrieves the batch row in the batch posting dialog that matches the given description.

        Args:
            description (str): The description to match against the batch posting dialog.
            amount (str): The amount to match against the batch posting dialog.

        Returns:
            None | BatchPostingRow: The batch row that matches the given description.

        Raises:
            NoMatchingBatchDescription: If no matching description is found.
            NextGenDuplicateImportException: If multiple matching descriptions are found.
        """
        div = self.desktop_app.dialog.child_window(auto_id="Data Area", control_type="Custom", found_index=0)
        tree_view = div.child_window(auto_id="ColScrollRegion: 0, RowScrollRegion: 0")
        batch_posting_rows = self.get_batch_posting_rows(tree_view)

        if batch_posting_rows is None:
            send_keys("%C")
            self.logger.debug("No matching batch description found")
            raise NoMatchingBatchDescriptionException(CONST.BATCH_DESCRIPTION_NOT_FOUND)

        rows = self._filter_batch_posting_rows(batch_posting_rows, trn, amount)
        rows_count = len(rows)
        if rows_count == 1:
            row: BatchPostingRow = rows[0]
            if not row.visible:
                self._click_down_n_times(row.index)
                row.visible = True
            with contextlib.suppress(_ctypes.COMError):
                row.edit_wrapper.click_input()
            self.logger.info("Batch row found")
            return row
        elif rows_count == 0:
            send_keys("%C")
            raise NoMatchingBatchDescriptionException(CONST.BATCH_DESCRIPTION_NOT_FOUND)
        else:
            send_keys("%C")
            self.logger.debug("More than one batch with the same description was found")
            raise NextGenDuplicateImportException("More than one batch with the same description was found")

    def close_batch_look_up_window(self) -> None:
        """Click close button in Batch Lookup Window."""
        self.logger.debug("Closing the Batch Lookup window")
        batch_lookup_window = self.desktop_app.dialog.child_window(title="Batch Lookup", control_type="Window")
        batch_lookup_window.set_focus()
        with contextlib.suppress(_ctypes.COMError):
            batch_lookup_window.child_window(title="Close", control_type="Button", found_index=0).click_input()

    @retry_if_pywin_error()  # type: ignore[misc]
    def get_batch_date(self) -> str:
        """This function gets the batch date.

        Returns:
            str: Batch date
        """
        batch_date = (
            self.window.child_window(auto_id="ColScrollRegion: 0, RowScrollRegion: 0")
            .child_window(control_type="DataItem")
            .child_window(title="Date", control_type="Edit")
            .get_value()
        )
        self.logger.debug(f"Batch date: {batch_date}")
        return batch_date

    def get_batch_id(self) -> str:
        """This function gets the batch ID in the batch lookup flow.

        Returns:
            str: Batch ID
        """
        batch_id = (
            self.window.child_window(auto_id="ColScrollRegion: 0, RowScrollRegion: 0")
            .child_window(control_type="DataItem")
            .child_window(title="ID", control_type="Edit")
            .get_value()
        )
        self.logger.debug(f"Batch ID: {batch_id}")
        return batch_id

    def enter_quick_batch_name(self, name: str) -> None:
        """The function enters a Bach name into a text field in the Quick Batch dialog window.

        Args:
            name (str): the name of the batch
        """
        self.logger.debug(f"Entering quick batch name: {name}")
        quick_batch_window = self.desktop_app.dialog.child_window(auto_id="NGEPMQuickBatch", control_type="Window")
        batch_name_field = quick_batch_window.child_window(auto_id="txtDesc", control_type="Edit")
        batch_name_field.wait("visible", 5)
        with contextlib.suppress(_ctypes.COMError):
            batch_name_field.set_edit_text(name)

    @retry_if_pywin_error(3, 3)  # type: ignore[misc]
    def press_l_to_open_ledger_in_dropdown(self) -> None:
        """Press "l" to open the ledger in the opened dropdown."""
        self.logger.debug("Pressing 'l' to open the ledger in the dropdown")
        send_keys("l")
        self.desktop_app.dialog.child_window(title="BatchLedger", control_type="Window").wait("visible", 10)

    def _open_dropdown_for_batch_row(self) -> None:
        """Click on the ledger left button to open the options dropdown for the selected row."""
        self.logger.debug("Opening the options dropdown for the selected row")
        button = self.desktop_app.dialog.child_window(title="cmdDrill", control_type="Button")
        with contextlib.suppress(_ctypes.COMError):
            button.click()
        time.sleep(2)



"""OpenCV module."""

import cv2
import numpy as np


class OpenCV:
    """OpenCV class for image processing."""

    @staticmethod
    def load_image(image_path: str) -> cv2.typing.MatLike:
        """Load image from path.

        Args:
            image_path: path of image

        Returns:
            image: loaded image
        """
        image = cv2.imread(image_path, cv2.IMREAD_GRAYSCALE)
        return image

    @staticmethod
    def is_strikethrough(img: np.ndarray) -> bool:
        """Detects whether text in an image is crossed out.

        Args:
            img (cv2.typing.MatLike): Image to check.

        Returns:
            bool: True if text is crossed out, False otherwise.
        """
        # Apply Gaussian blur to reduce noise
        blurred = cv2.GaussianBlur(img, (5, 5), 0)

        # Apply adaptive thresholding to obtain a binary image
        thresh = cv2.adaptiveThreshold(blurred, 255, cv2.ADAPTIVE_THRESH_MEAN_C, cv2.THRESH_BINARY_INV, 11, 2)

        # Find horizontal lines using morphological operations
        horizontal_kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (25, 1))
        horizontal_lines = cv2.morphologyEx(thresh, cv2.MORPH_OPEN, horizontal_kernel, iterations=2)

        # Find the contours of the horizontal lines
        contours, hierarchy = cv2.findContours(horizontal_lines, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

        # Analyze the contours to detect a strikethrough line
        for cnt in contours:
            x, y, w, h = cv2.boundingRect(cnt)
            # Check if the line is long enough and has a reasonable width
            if w > img.shape[1] * 0.65 and h < img.shape[0] * 0.3:
                return True
        return False

    @staticmethod
    def check_strikethrough_in_text_image(image_path: str) -> bool:
        """Method to load image and check if it has strikethrough.

        Args:
            image_path: path of image

        Returns:
            bool: True if strikethrough is present, False otherwise
        """
        image = OpenCV.load_image(image_path)
        return OpenCV.is_strikethrough(image)



"""Container for all values used in the automation."""
from enum import Enum


class _Constants:
    """Container for all values used in the automation."""

    BATCH_DESCRIPTION_NOT_FOUND = "A batch with the provided description was not found"

    class WORK_ITEM_PHASE(Enum):
        NOT_STARTED = "NOT_STARTED"
        IN_PROGRESS = "IN_PROGRESS"
        RETRY = "RETRY"
        IMPORTED = "IMPORTED"
        START_IMPORTING = "START_IMPORTING"


CONST = _Constants()



"""All config variables goes here."""


class Config:
    """Static class container for all variables."""

    class DIRECTORIES:
        """Directories to interact with NextGen."""

        BASE_DIR = r"C:/NextGen"
        FILE_MAINTENANCE = r"C:/NextGen/NGFileMaint.exe"
        DOC_MANAGEMENT = r"C:/NextGen/NextGenICS.exe"
        ENTERPRISE_PM = r"C:/NextGen/NextGenEPM.exe"

    class DATABASES:
        """NextGen available databases."""

        TEST = "NGTEST"
        PROD = "NGPROD"



"""ImportFile module."""

from retry import retry
from t_nextgen.nextgen_window import NextGenWindow

import contextlib
import _ctypes
from t_nextgen.exceptions import (
    NextGenEdiNotFoundException,
    NextGenControlLostException,
    NextGenDuplicateImportException,
    NextGenImportTimeoutException,
    NextGenPatientPayException,
    PostingSheetTimeOutException,
)
from t_desktop.config import IS_WINDOWS_OS

if IS_WINDOWS_OS:
    from pywinauto.application import WindowSpecification
    from pywinauto.findwindows import ElementNotFoundError
    from pywinauto.keyboard import send_keys
    from t_desktop.decorators import capture_screenshot_if_exception
import time
from datetime import datetime, timedelta

from pathlib import Path
from t_nextgen.utils import utility


class ImportFileWindow(NextGenWindow):
    """ImportFile class with methods to interact with import file window."""

    @property
    def window(self) -> WindowSpecification:
        """Get control of Import Window."""
        self.desktop_app.dialog.child_window(title="frmERA", control_type="Window").wait("exists", timeout=10)
        return self.desktop_app.dialog.child_window(title="frmERA", control_type="Window")

    @property
    def enhanced_report_mode_window(self) -> WindowSpecification:
        """Get control of Enhanced Report Mode Window."""
        return self.desktop_app.dialog.child_window(
            title="ERA Import Posting - Enhanced Report Mode", control_type="Window"
        )

    @property
    def report_viewer_window(self) -> WindowSpecification:
        """Return the report_viewer_window window element."""
        return self.desktop_app.dialog.child_window(auto_id="ReportViewer", control_type="Window")

    def click_era_posting_from_processes_menu(self) -> None:
        """Click on the ERA Posting option in the Processes menu."""
        self.logger.debug("Clicking on ERA Posting from Processes Menu.")
        self.desktop_app.click_file_menu_button()
        file_menu = self.desktop_app.get_file_menu()
        self.desktop_app.click_processes_button_in_file_menu(file_menu)
        processes_menu = self.desktop_app.get_processes_menu_from_file_menu(file_menu)
        with contextlib.suppress(_ctypes.COMError):
            processes_menu.child_window(title="ERA Posting...", control_type="MenuItem", found_index=0).click_input()

    def click_cancel_import_window(self) -> None:
        """Click on cancel button in Import Window."""
        self.logger.debug("Clicking on Cancel button in Import Window.")
        with contextlib.suppress(_ctypes.COMError):
            self.window.child_window(title="cmdCancel", auto_id="cmdCancel", control_type="Button").click()

    @retry(tries=3, delay=1)
    def select_directory_import(self) -> None:
        """Select the directory import option."""
        self.logger.debug("Selecting Directory Import option.")
        try:
            with contextlib.suppress(_ctypes.COMError):
                self.window.child_window(
                    title="optFolderSearch", auto_id="optFolderSearch", control_type="RadioButton"
                ).click()
        except TimeoutError as e:
            self.logger.warning("Directory Import option not found within timeout.")
            self.click_cancel_import_window()
            self.click_era_posting_from_processes_menu()
            raise e

    def set_835_directory_path(self, folder_835_path: str) -> None:
        """Set the path for the 835 directory.

        Args:
            folder_835_path (str): the folder path where 835 files are stored
        """
        self.logger.debug(f"Setting 835 Directory Path: {folder_835_path}")
        with contextlib.suppress(_ctypes.COMError):
            self.window.child_window(title="txtFolderName", auto_id="txtFolderName", control_type="Edit").set_text(
                folder_835_path
            )

    def click_find_button(self) -> None:
        """Click on find button."""
        self.logger.debug("Clicking on Find button.")
        with contextlib.suppress(_ctypes.COMError):
            self.window.child_window(title="cmdFind", auto_id="cmdFind", control_type="Button").click()

    def get_edi_pane(self) -> WindowSpecification:
        """Get EDI Pane control.

        Returns:
            WindowSpecification: EDI Pane Control
        """
        pane = self.window.child_window(auto_id="frmFolderResults", control_type="Group").child_window(
            auto_id="lstListing", control_type="Pane"
        )
        pane.wait("exists", timeout=15)
        return pane

    @capture_screenshot_if_exception()
    def get_edi_checkbox_count(self, edi_pane: WindowSpecification) -> int:
        """Get the EDI CheckBox count aka total number of EDIs to be imported.

        Args:
            edi_pane (WindowSpecification): EDI Pane control

        Returns:
            int: number of EDIs found
        """
        check_box_count = len(edi_pane.descendants(control_type="CheckBox"))
        if check_box_count == 0:
            with contextlib.suppress(_ctypes.COMError):
                self.desktop_app.dialog.child_window(title="cmdCancel", control_type="Button").click()
            self.logger.debug("No EDI files found in the 835 folder.")
            raise NextGenEdiNotFoundException("No EDI files found in the 835 folder")
        return check_box_count

    def select_edi_row_checkboxes(self) -> None:
        """Select all the EDIs as visible in ERA processing window in NextGen App."""
        edi_pane = self.get_edi_pane()
        check_box_count = self.get_edi_checkbox_count(edi_pane)
        self.logger.debug(f"Selecting EDI Row Checkboxes. {check_box_count} EDIs found.")
        for i in range(check_box_count):
            send_keys("{DOWN}")
            time.sleep(1)
            with contextlib.suppress(_ctypes.COMError):
                checkbox_index = 6 if i > 6 else i
                edi_pane.child_window(control_type="CheckBox", found_index=checkbox_index).click()

    def import_edi_files(self) -> None:
        """Click on import button in the ERA processing window in NextGen App."""
        self.logger.debug("Clicking on Import button.")
        data = self.window.child_window(title="cmdImport", auto_id="cmdImport", control_type="Button")
        with contextlib.suppress(_ctypes.COMError):
            data.click()
        self.wait_for_edi_import_to_finish()

    def wait_for_edi_import_to_finish(self) -> None:
        """Waits for the EDIs to be imported into the NextGen App."""
        self.logger.debug("Waiting for EDI import to finish.")
        start_time = datetime.now()
        end_time = start_time + timedelta(seconds=3600)
        time.sleep(1)
        max_tries = 7
        current_try = 0
        try:
            if self.patient_has_duplicated_import():
                self.logger.debug("Patient has duplicated import.")
                raise NextGenDuplicateImportException("Current 835 folder has EDI that is already imported")
            elif self.patient_pay_file_modal_exist():
                self.logger.debug(
                    "There are more than 835 files in the folder. Please ensure only 835 files remain in the folder."
                )
                raise NextGenPatientPayException("Remove Patient Pay from the current 835 Folder")
        except RuntimeError:
            raise NextGenControlLostException("Lost control of NextGen Application")

        while self.import_edi_modal_exist() and datetime.now() < end_time:
            if current_try > max_tries:
                raise Exception(
                    "Could not import current payment folder or lost NextGen reference and could not recover"
                )
            try:
                self.app = self.desktop_app.connect_to_logged_in_screen()
                self._process_import_modal()
            except (ElementNotFoundError, TimeoutError):
                current_try += 1
                time.sleep(10)

        if datetime.now() >= end_time:
            raise NextGenImportTimeoutException("Could not import current payment folder within 3600 seconds")

    def _process_import_modal(self) -> None:
        """Handle Import Processing Dialog."""
        modal = self.desktop_app.dialog.child_window(title="NextGen", control_type="Window")
        if modal.exists(timeout=3, retry_interval=1):
            buttons_list = modal.wrapper_object().descendants(control_type="Button")
            if len(buttons_list) == 2 and buttons_list[0].texts()[0] == "OK" and buttons_list[1].texts()[0] == "Close":
                with contextlib.suppress(_ctypes.COMError):
                    modal.child_window(title="OK", auto_id="2", control_type="Button").click()

    def convert_edi_files_to_excel(self, excel_file_path: str) -> None:
        """Converts EDI files to Excel format by interacting with the application's excel export functionality.

        Args:
            excel_file_path (str): The file path where the exported excel should be saved.
        """
        self.logger.debug("Converting EDI files to Excel.")
        self.click_export_to_excel_icon()
        self.set_excel_export_path(excel_file_path)
        self.export_edi_files()
        pathlib_path = Path(rf"{excel_file_path}")
        try:
            utility.wait_until_file_downloads(pathlib_path.parent, pathlib_path.name, wait_time=500)
        except TimeoutError:
            if self.is_export_modal_open():
                self.desktop_app.dialog.child_window(auto_id="Cancel", control_type="Button")
                raise PostingSheetTimeOutException("PostingSheet not created within 500 seconds")

    def convert_edi_files_to_csv(self, csv_file_path: str) -> None:
        """Converts EDI files to CSV format by interacting with the application's csv export functionality.

        Args:
            csv_file_path (str): The file path where the exported csv should be saved.
        """
        self.logger.debug("Converting EDI files to CSV.")
        self.click_export_to_csv_icon()
        self.set_csv_export_path(csv_file_path)
        self.export_edi_files()
        pathlib_path = Path(rf"{csv_file_path}")
        try:
            utility.wait_until_file_downloads(pathlib_path.parent, pathlib_path.name, wait_time=500)
        except TimeoutError:
            if self.is_export_modal_open():
                self.desktop_app.dialog.child_window(auto_id="Cancel", control_type="Button")
                raise PostingSheetTimeOutException("PostingSheet not created within 500 seconds")

    def is_export_modal_open(self) -> bool:
        """Checks if the export modal dialog is currently open in the NextGen."""
        try:
            with contextlib.suppress(_ctypes.COMError):
                return self.desktop_app.dialog.child_window(auto_id="FileDialog", control_type="Custom").exists(
                    timeout=3, retry_interval=1
                )
        except ElementNotFoundError:
            return False

    def export_edi_files(self) -> None:
        """Click the export button in the window after clicking the excel icon in the NextGen App."""
        self.logger.debug("Exporting EDI files to Excel.")
        with contextlib.suppress(_ctypes.COMError):
            self.report_viewer_window.child_window(title="Export", control_type="Button").click_input()

    def set_excel_export_path(self, export_path: str) -> None:
        """Set csv export path for converted EDIs in the window that pops up after clicking the excel icon.

        Args:
            export_path (str): The path where the generated csv would be stored
        """
        self.logger.debug(f"Setting Excel Export Path: {export_path}")
        closest_window = self.desktop_app.dialog.child_window(
            auto_id="ReportViewer", control_type="Window"
        ).child_window(auto_id="FileDialog", control_type="Custom")
        with contextlib.suppress(_ctypes.COMError):
            closest_window.child_window(control_type="Edit").set_text(export_path)

    def click_export_to_excel_icon(self) -> None:
        """Convert imported EDIs to Excel. Clicks on the Export to Excel icon in the NextGen App."""
        self.logger.debug("Clicking on Export to Excel icon.")
        excel_button = (
            self.desktop_app.dialog.child_window(auto_id="ReportViewer", control_type="Window")
            .child_window(control_type="ToolBar", class_name="ToolBar")
            .child_window(control_type="Button", found_index=7)
        )
        if excel_button.exists(timeout=5):
            with contextlib.suppress(_ctypes.COMError):
                excel_button.click_input()
                self.logger.info("Excel button clicked successfully.")

            closest_window = self.desktop_app.dialog.child_window(
                auto_id="ReportViewer", control_type="Window"
            ).child_window(auto_id="FileDialog", control_type="Custom")
            closest_window.wait("ready", timeout=3)
            self.logger.info("FileDialog window is ready.")
        else:
            self.logger.warning("Excel button not found within timeout.")

    def click_export_to_csv_icon(self) -> None:
        """Convert imported EDIs to Excel. Clicks on the Export to Excel icon in the NextGen App."""
        self.logger.debug("Clicking on Export to Excel icon.")
        csv_button = (
            self.desktop_app.dialog.child_window(auto_id="ReportViewer", control_type="Window")
            .child_window(control_type="ToolBar", class_name="ToolBar")
            .child_window(control_type="Button", found_index=8)
        )
        if csv_button.exists(timeout=5):
            with contextlib.suppress(_ctypes.COMError):
                csv_button.click_input()
                self.logger.info("Excel button clicked successfully.")

            closest_window = self.desktop_app.dialog.child_window(
                auto_id="ReportViewer", control_type="Window"
            ).child_window(auto_id="FileDialog", control_type="Custom")
            closest_window.wait("ready", timeout=3)
            self.logger.info("FileDialog window is ready.")
        else:
            self.logger.warning("Excel button not found within timeout.")

    def set_csv_export_path(self, export_path: str) -> None:
        """Set csv export path for converted EDIs in the window that pops up after clicking the excel icon.

        Args:
            export_path (str): The path where the generated csv would be stored
        """
        self.logger.debug(f"Setting Excel Export Path: {export_path}")
        closest_window = self.desktop_app.dialog.child_window(
            auto_id="ReportViewer", control_type="Window"
        ).child_window(auto_id="FileDialog", control_type="Custom")
        with contextlib.suppress(_ctypes.COMError):
            closest_window.child_window(control_type="Edit").set_text(export_path)

    def import_edi_modal_exist(self) -> bool:
        """Look at a reference to a modal on the next screen to see if the import is finished.

        The reference of this template comes from the screen after importation.
        If it finds the screen, it returns False due to negation; if it doesn't find the screen,
        a _ctypes.COMError error is raised, which falls into the exception and returns True to
        maintain waiting for the download.
        """
        with contextlib.suppress((_ctypes.COMError, TimeoutError, RuntimeError)):
            return not self.desktop_app.dialog.child_window(
                title="ERA Import Posting - Enhanced Report Mode", auto_id="ReportViewer", control_type="Window"
            ).exists(timeout=5, retry_interval=1)
            # Return True if a timeouterror or RuntimeError happens to prevent break the import
        return True

    def patient_has_duplicated_import(self) -> bool:
        """This method checks if the patient has duplicated import.

        Args:
            dialog2 (WindowSpecification): dialog2 window
        Returns:
            bool: True if the patient has duplicated import, False otherwise
        """
        with contextlib.suppress(_ctypes.COMError):
            return self.desktop_app.dialog.child_window(
                auto_id="65535",
                control_type="Text",
                title="Some files/checks were identified as duplicate imports.  Continue?",
            ).exists(timeout=3, retry_interval=1)
        return False

    def _handle_duplicate_import(self) -> None:
        """Handle Duplicate Import Dialog.

        Raises:
            NextGenDuplicateImportException: For duplicate import
        """
        nextgen_dialog = self.desktop_app.dialog.child_window(title="NextGen", control_type="Window")
        with contextlib.suppress(_ctypes.COMError):
            nextgen_dialog.child_window(title="No", auto_id="7", control_type="Button").click_input()
            self.window.child_window(title="cmdCancel", control_type="Button").click_input()

    def _handle_patient_pay_exception(self) -> None:
        """Handles Patient Pay Dialog."""
        with contextlib.suppress(_ctypes.COMError):
            self.desktop_app.dialog.child_window(title="Close", control_type="Button", found_index=0).click()
            self.desktop_app.dialog.child_window(title="cmdCancel", control_type="Button").click()

    def patient_pay_file_modal_exist(self) -> bool:
        """This method checks if the Patient Pay file modal exists.

        Returns:
            bool: True if the Patient Pay file modal exists, False otherwise
        """
        with contextlib.suppress(_ctypes.COMError):
            return self.desktop_app.dialog.child_window(
                auto_id="65535",
                control_type="Text",
                title="Please select either ERA files or Patient Pay files but not both.",
            ).exists(timeout=3, retry_interval=1)
        return False

    def click_columns_button(self) -> None:
        """Click on the Columns button in the Import Window."""
        self.logger.debug("Clicking on Columns button in Import Window.")
        with contextlib.suppress(_ctypes.COMError):
            self.enhanced_report_mode_window.child_window(title="Columns", control_type="Button").click_input()

    def click_totals_button(self) -> None:
        """Click on the Totals button in the Import Window."""
        self.logger.debug("Clicking on Totals button in Import Window.")
        with contextlib.suppress(_ctypes.COMError):
            self.enhanced_report_mode_window.child_window(title="Totals", control_type="Button").click_input()

    def click_options_button(self) -> None:
        """Click on the Options button in the Import Window."""
        self.logger.debug("Clicking on Options button in Import Window.")
        with contextlib.suppress(_ctypes.COMError):
            self.enhanced_report_mode_window.child_window(title="Options", control_type="Button").click_input()

    def click_header_footer_button(self) -> None:
        """Click on the Header/Footer button in the Import Window."""
        self.logger.debug("Clicking on Header/Footer button in Import Window.")
        with contextlib.suppress(_ctypes.COMError):
            self.enhanced_report_mode_window.child_window(title="Header/Footer", control_type="Button").click_input()

    def click_sorting_button(self) -> None:
        """Click on the Sorting button in the Import Window."""
        self.logger.debug("Clicking on Sorting button in Import Window.")
        with contextlib.suppress(_ctypes.COMError):
            self.enhanced_report_mode_window.child_window(title="Sorting", control_type="Button").click_input()

    def uncheck_all_group_by_checkboxes(self, max_iterations: int = 10) -> None:
        """Uncheck all checkboxes in the Group By Window.

        Args:
            max_iterations (int): Maximum number of iterations to uncheck all checkboxes
        """
        self.logger.debug("Unchecking all group by checkboxes in the Sorting Window.")
        iteration = 0
        while iteration < max_iterations:
            sorting = self.enhanced_report_mode_window.child_window(auto_id="Sorting", control_type="Custom")
            data_grid = sorting.child_window(control_type="DataGrid", found_index=0)
            data_items = data_grid.descendants(control_type="DataItem")
            self.logger.debug(len(data_items))
            for data_item in data_items:
                if data_item.is_visible():
                    text = data_item.descendants(control_type="Text")[0].window_text()
                    if text == "CSC Claim Payment":
                        self.unselect_check_box(data_item)
                        return
                    self.unselect_check_box(data_item)
                    with contextlib.suppress(_ctypes.COMError):
                        data_item.descendants(control_type="Custom")[0].click_input()
            send_keys("{PGDN}")

    def click_column_name_check_box(self) -> None:
        """Click on the Column Name checkbox in the Columns Window."""
        self.logger.debug("Clicking on the Column Name checkbox in the Columns Window.")
        s_text = self.enhanced_report_mode_window.child_window(auto_id="FilterText", control_type="Edit")
        with contextlib.suppress(_ctypes.COMError):
            s_text.click_input()
        send_keys("{TAB}")
        send_keys("{TAB}")
        send_keys("{SPACE}")

    def unselect_check_box(self, data_item: WindowSpecification, index: int = 2) -> None:
        """Click on the checkbox in the ImportFile template options Window.

        Args:
            data_item (WindowSpecification): DataItem control
            index (int): Index of the checkbox in the DataItem control
        """
        check_box = data_item.descendants(control_type="CheckBox")[index]
        if check_box.get_toggle_state() == 1:
            check_box.toggle()

    def unselect_count_check_boxes(self, max_iterations: int = 10):
        """Unselect the Count checkboxes in the Columns Window.

        Args:
            max_iterations (int): Maximum number of iterations to unselect the Count checkboxes
        """
        iteration = 0
        while iteration < max_iterations:
            totals = self.enhanced_report_mode_window.child_window(auto_id="Totals", control_type="Custom")
            data_grid = totals.child_window(control_type="DataGrid", found_index=0)
            data_items = data_grid.descendants(control_type="DataItem")
            self.logger.debug(f"Number of data items: {len(data_items)}")
            for data_item in data_items:
                if data_item.is_visible():
                    text = data_item.descendants(control_type="Text")[0].window_text()
                    if text == "CSC Claim Payment":
                        self.unselect_check_box(data_item)
                        return
                    self.unselect_check_box(data_item)
                    with contextlib.suppress(_ctypes.COMError):
                        data_item.descendants(control_type="Custom")[0].click_input()
            send_keys("{PGDN}")

    def unselect_count_records_check_box(self) -> None:
        """Unselect the Count Records checkbox in the Columns Window."""
        self.logger.debug("Unselecting the Count Records checkbox in the Columns Window.")
        count_record_check_box = self.enhanced_report_mode_window.child_window(
            title="Count Records", control_type="CheckBox"
        )
        if count_record_check_box.get_toggle_state() == 1:
            count_record_check_box.toggle()

    def click_refresh_report_button(self):
        """Click on the Refresh Report button."""
        self.logger.debug("Clicking on the Refresh Report button.")
        custom = self.enhanced_report_mode_window.child_window(auto_id="Me", control_type="Custom")
        buttons = custom.descendants(control_type="Button")
        for button in buttons:
            if button.descendants(control_type="Text")[0].window_text() == "_Refresh Report":
                with contextlib.suppress(_ctypes.COMError):
                    button.click_input()
                break

    def unselect_column_text_wrapping_check_box(self) -> None:
        """Unselect the Column Text Wrapping checkbox in the Columns Window."""
        self.logger.debug("Unselecting the Column Text Wrapping checkbox in the Columns Window.")
        check_box = self.enhanced_report_mode_window.child_window(title="Column Text Wrapping", control_type="CheckBox")
        if check_box.get_toggle_state() == 1:
            check_box.toggle()

    def unselect_all_check_box_from_header(self) -> None:
        """Unselect all checkboxes from the Header Window."""
        self.logger.debug("Unselecting all checkboxes from the Header Window.")

        header = self.enhanced_report_mode_window.child_window(auto_id="HeaderFooter", control_type="Custom")
        check_boxes = header.descendants(control_type="CheckBox")
        for check_box in check_boxes:
            if check_box.get_toggle_state() == 1:
                check_box.toggle()

    def click_close_button_in_import_posting_report_window(self):
        """Click on the Close button in the Import Posting Report Window."""
        self.logger.debug("Clicking on the Close button in the Import Posting Report Window.")
        with contextlib.suppress(_ctypes.COMError):
            self.report_viewer_window.child_window(control_type="ToolBar", class_name="ToolBar").child_window(
                control_type="Button", found_index=18
            ).click_input()
